# sage_setup: distribution = sagemath-combinat
# sage.doctest: needs sage.combinat sage.groups
r"""
Integer vectors modulo the action of a permutation group

AUTHORS:

* Nicolas Borie (2010-2012) - original module
* Jukka Kohonen (2023) - fast cardinality method, :issue:`36787`, :issue:`36681`
"""
# ****************************************************************************
#    Copyright (C) 2010-12 Nicolas Borie <nicolas.borie at math dot u-psud.fr>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#              The full text of the GPL is available at:
#                    https://www.gnu.org/licenses/
# ****************************************************************************

from sage.structure.unique_representation import UniqueRepresentation
from sage.rings.semirings.non_negative_integer_semiring import NN

from sage.categories.infinite_enumerated_sets import InfiniteEnumeratedSets
from sage.categories.finite_enumerated_sets import FiniteEnumeratedSets

from sage.structure.list_clone import ClonableIntArray
from sage.sets.recursively_enumerated_set import RecursivelyEnumeratedSet_forest

from sage.combinat.enumeration_mod_permgroup import is_canonical, orbit, canonical_children, canonical_representative_of_orbit_of

from sage.combinat.integer_vector import IntegerVectors

from sage.rings.power_series_ring import PowerSeriesRing
from sage.rings.rational_field import QQ
from sage.rings.integer import Integer
from sage.misc.misc_c import prod
from sage.arith.misc import binomial


class IntegerVectorsModPermutationGroup(UniqueRepresentation):
    r"""
    Return an enumerated set containing integer vectors which are
    maximal in their orbit under the action of the permutation group
    ``G`` for the lexicographic order.

    In Sage, a permutation group `G` is viewed as a subgroup of the
    symmetric group `S_n` of degree `n` and `n` is said to be the degree
    of `G`.  Any integer vector `v` is said to be canonical if it
    is maximal in its orbit under the action of `G`. `v` is
    canonical if and only if

    .. MATH::

        v = \max_{\text{lex order}} \{g \cdot v | g \in G \}

    The action of `G` is on position. This means for example that the
    simple transposition `s_1 = (1, 2)` swaps the first and the second
    entries of any integer vector `v = [a_1, a_2, a_3, \dots , a_n]`

    .. MATH::

        s_1 \cdot v = [a_2, a_1, a_3, \dots , a_n]

    This function returns a parent which contains, from each orbit
    orbit under the action of the permutation group `G`, a single
    canonical vector.  The canonical vector is the one that is maximal
    within the orbit according to lexicographic order.

    INPUT:

    - ``G`` -- a permutation group
    - ``sum`` -- (default: ``None``) - a nonnegative integer
    - ``max_part`` -- (default: ``None``) - a nonnegative integer setting the
      maximum value for every element
    - ``sgs`` -- (default: ``None``) - a strong generating system of the
      group `G`. If you do not provide it, it will be calculated at the
      creation of the parent

    OUTPUT:

    - If ``sum`` and ``max_part`` are None, it returns the infinite
      enumerated set of all integer vectors (lists of integers) maximal
      in their orbit for the lexicographic order.  Exceptionally, if
      the domain of ``G`` is empty, the result is a finite enumerated
      set that contains one element, namely the empty vector.

    - If ``sum`` is an integer, it returns a finite enumerated set
      containing all integer vectors maximal in their orbit for the
      lexicographic order and whose entries sum to ``sum``.

    EXAMPLES:

    Here is the set enumerating integer vectors modulo the action of the cyclic
    group of `3` elements::

        sage: I = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3)]]))
        sage: I.category()
        Category of infinite enumerated quotients of sets
        sage: I.cardinality()
        +Infinity
        sage: I.list()
        Traceback (most recent call last):
        ...
        NotImplementedError: cannot list an infinite set
        sage: p = iter(I)
        sage: for i in range(10): next(p)
        [0, 0, 0]
        [1, 0, 0]
        [2, 0, 0]
        [1, 1, 0]
        [3, 0, 0]
        [2, 1, 0]
        [2, 0, 1]
        [1, 1, 1]
        [4, 0, 0]
        [3, 1, 0]

    The method
    :meth:`~sage.combinat.integer_vectors_mod_permgroup.IntegerVectorsModPermutationGroup_All.is_canonical`
    tests if an integer vector is maximal in its orbit. This method
    is also used in the containment test::

        sage: I = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3,4)]]))
        sage: I.is_canonical([5,2,0,4])
        True
        sage: I.is_canonical([5,0,6,4])
        False
        sage: I.is_canonical([1,1,1,1])
        True
        sage: [2,3,1,0] in I
        False
        sage: [5,0,5,0] in I
        True
        sage: 'Bla' in I
        False
        sage: I.is_canonical('bla')
        Traceback (most recent call last):
        ...
        AssertionError: bla should be a list or an integer vector

    If you give a value to the extra argument ``sum``, the set returned
    will be a finite set containing only canonical vectors whose entries
    sum to ``sum``.::

        sage: I = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3)]]), sum=6)
        sage: I.cardinality()
        10
        sage: I.list()
        [[6, 0, 0], [5, 1, 0], [5, 0, 1], [4, 2, 0], [4, 1, 1],
         [4, 0, 2], [3, 3, 0], [3, 2, 1], [3, 1, 2], [2, 2, 2]]
        sage: I.category()
        Join of Category of finite enumerated sets
            and Category of subquotients of finite sets
            and Category of quotients of sets

    To get the orbit of any integer vector `v` under the action of the group,
    use the method :meth:`~sage.combinat.integer_vectors_mod_permgroup.IntegerVectorsModPermutationGroup_All.orbit`;
    we convert the returned set of vectors into a list in increasing lexicographic order,
    to get a reproducible test::

        sage: sorted(I.orbit([6,0,0]))
        [[0, 0, 6], [0, 6, 0], [6, 0, 0]]
        sage: sorted(I.orbit([5,1,0]))
        [[0, 5, 1], [1, 0, 5], [5, 1, 0]]
        sage: I.orbit([2,2,2])
        {[2, 2, 2]}

    Even without constraints, for an empty domain the result is
    a singleton set::

        sage: G = PermutationGroup([], domain=[])
        sage: sgs = tuple(tuple(s) for s in G.strong_generating_system())
        sage: list(IntegerVectorsModPermutationGroup(G, sgs=sgs))
        [[]]


    .. WARNING::

        Because of :issue:`36527`, permutation groups that have
        different domains but similar generators can be erroneously
        treated as the same group.  This will silently produce
        erroneous results.  To avoid this issue, compute a strong
        generating system for the group as::

            sgs = tuple(tuple(s) for s in G.strong_generating_system())

        and provide it as the optional ``sgs`` argument to the
        constructor.

    TESTS:

    Let us check that canonical integer vectors of the symmetric group
    are just nonincreasing lists of integers::

        sage: I = IntegerVectorsModPermutationGroup(SymmetricGroup(5))  # long time
        sage: p = iter(I)                                               # long time
        sage: for i in range(100):                                      # long time
        ....:     v = list(next(p))
        ....:     assert sorted(v, reverse=True) == v

    We now check that there are as many canonical vectors under the
    symmetric group `S_n` whose entries sum to `d` as there are
    partitions of `d` of at most `n` parts::

        sage: I = IntegerVectorsModPermutationGroup(SymmetricGroup(5))  # long time
        sage: for i in range(10):                                       # long time
        ....:     d1 = I.subset(i).cardinality()
        ....:     d2 = Partitions(i, max_length=5).cardinality()
        ....:     print(d1)
        ....:     assert d1 == d2
        1
        1
        2
        3
        5
        7
        10
        13
        18
        23

    Another corner case is trivial groups. For the trivial group ``G``
    acting on a list of length `n`, all integer vectors of length `n`
    are canonical::

        sage: # long time
        sage: G = PermutationGroup([[(6,)]])
        sage: G.cardinality()
        1
        sage: sgs = tuple(tuple(s) for s in G.strong_generating_system())
        sage: I = IntegerVectorsModPermutationGroup(G, sgs=sgs)
        sage: for i in range(10):
        ....:     d1 = I.subset(i).cardinality()
        ....:     d2 = IntegerVectors(i,6).cardinality()
        ....:     print(d1)
        ....:     assert d1 == d2
        1
        6
        21
        56
        126
        252
        462
        792
        1287
        2002
    """
    @staticmethod
    def __classcall__(cls, G, sum=None, max_part=None, sgs=None):
        r"""
        TESTS::

            sage: I = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3)]]))
            sage: I = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3)]]), None)
            sage: I = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3)]]), 2)
            sage: I = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3)]]), -2)
            Traceback (most recent call last):
            ...
            ValueError: Value -2 in not in Non negative integer semiring.
            sage: I = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3)]]), 8, max_part=5)
        """
        if sum is None and max_part is None:
            # No constraints.
            if G.domain():
                # Nonempty domain, infinite set.
                return IntegerVectorsModPermutationGroup_All(G, sgs=sgs)
            else:
                # Empty domain, singleton set.
                return IntegerVectorsModPermutationGroup_with_constraints(
                    G, 0, max_part=-1, sgs=sgs)
        else:
            # Some constraints, either sum or max_part or both.
            if sum is not None:
                assert sum == NN(sum)
            if max_part is not None:
                assert max_part == NN(max_part)
            return IntegerVectorsModPermutationGroup_with_constraints(
                G, sum, max_part, sgs=sgs)


class IntegerVectorsModPermutationGroup_All(UniqueRepresentation, RecursivelyEnumeratedSet_forest):
    r"""
    A class for integer vectors enumerated up to the action of a
    permutation group.

    A Sage permutation group is viewed as a subgroup of the symmetric
    group `S_n` for a certain `n`. This group has a natural action by
    position on vectors of length `n`. This class implements a set
    which keeps a single vector for each orbit. We say that a vector
    is canonical if it is the maximum in its orbit under the action of
    the permutation group for the lexicographic order.

    EXAMPLES::

        sage: I = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3,4)]])); I
        Integer vectors of length 4 enumerated up to the action of
         Permutation Group with generators [(1,2,3,4)]
        sage: I.cardinality()
        +Infinity
        sage: TestSuite(I).run()
        sage: it = iter(I)
        sage: [next(it), next(it), next(it), next(it), next(it)]
        [[0, 0, 0, 0],
         [1, 0, 0, 0],
         [2, 0, 0, 0],
         [1, 1, 0, 0],
         [1, 0, 1, 0]]
        sage: x = next(it); x
        [3, 0, 0, 0]
        sage: I.first()
        [0, 0, 0, 0]

    TESTS::

        sage: TestSuite(I).run()
    """

    def __init__(self, G, sgs=None):
        """
        TESTS::

            sage: I = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3,4)]]))
            sage: I
            Integer vectors of length 4 enumerated up to the action of
             Permutation Group with generators [(1,2,3,4)]
            sage: I.category()
            Category of infinite enumerated quotients of sets
            sage: TestSuite(I).run()
        """
        RecursivelyEnumeratedSet_forest.__init__(self, algorithm='breadth', category=InfiniteEnumeratedSets().Quotients())
        self._permgroup = G
        self.n = G.degree()

        # self.sgs: strong_generating_system
        if sgs is None:
            self._sgs = G.strong_generating_system()
        else:
            self._sgs = [list(x) for x in list(sgs)]

    def _repr_(self):
        """
        TESTS::

            sage: IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3)]]))
            Integer vectors of length 3 enumerated up to the action of Permutation Group with generators [(1,2,3)]
        """
        return "Integer vectors of length %s enumerated up to the action of %r" % (self.n, self._permgroup)

    def ambient(self):
        r"""
        Return the ambient space from which ``self`` is a quotient.

        EXAMPLES::

            sage: S = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3,4)]]))
            sage: S.ambient()
            Integer vectors of length 4
        """
        return IntegerVectors(length=self.n)

    def lift(self, elt):
        r"""
        Lift the element ``elt`` inside the ambient space from which ``self`` is a quotient.

        EXAMPLES::

            sage: S = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3,4)]]))
            sage: v = S.lift(S([4,3,0,1])); v
            [4, 3, 0, 1]
            sage: type(v)
            <class 'list'>
        """
        # TODO: For now, Sage integer vectors are just python list.
        # Once Integer vectors will have an element class, update this
        # code properly
        return list(elt)

    def retract(self, elt):
        r"""
        Return the canonical representative of the orbit of the
        integer ``elt`` under the action of the permutation group
        defining ``self``.

        If the element ``elt`` is already maximal in its orbit for
        the lexicographic order, ``elt`` is thus the good
        representative for its orbit.

        EXAMPLES::

            sage: [0,0,0,0] in IntegerVectors(0,4)
            True
            sage: [1,0,0,0] in IntegerVectors(1,4)
            True
            sage: [0,1,0,0] in IntegerVectors(1,4)
            True
            sage: [1,0,1,0] in IntegerVectors(2,4)
            True
            sage: [0,1,0,1] in IntegerVectors(2,4)
            True
            sage: S = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3,4)]]))
            sage: S.retract([0,0,0,0])
            [0, 0, 0, 0]
            sage: S.retract([1,0,0,0])
            [1, 0, 0, 0]
            sage: S.retract([0,1,0,0])
            [1, 0, 0, 0]
            sage: S.retract([1,0,1,0])
            [1, 0, 1, 0]
            sage: S.retract([0,1,0,1])
            [1, 0, 1, 0]
        """
        # TODO: Once Sage integer vector will have a data structure
        # based on ClonableIntArray, remove the conversion intarray
        assert len(elt) == self.n, "%s is a quotient set of %s" % (self, self.ambient())
        intarray = self.element_class(self, elt, check=False)
        return self.element_class(self, canonical_representative_of_orbit_of(self._sgs, intarray), check=False)

    def roots(self):
        r"""
        Return the root of generation of ``self``. This method is
        required to build the tree structure of ``self`` which
        inherits from the class :class:`~sage.sets.recursively_enumerated_set.RecursivelyEnumeratedSet_forest`.

        EXAMPLES::

            sage: I = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3,4)]]))
            sage: I.roots()
            [[0, 0, 0, 0]]
        """
        return [self.element_class(self, self.n*[0,], check=False)]

    def children(self, x):
        r"""
        Return the list of children of the element ``x``. This method
        is required to build the tree structure of ``self`` which
        inherits from the class :class:`~sage.sets.recursively_enumerated_set.RecursivelyEnumeratedSet_forest`.

        EXAMPLES::

            sage: I = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3,4)]]))
            sage: I.children(I([2,1,0,0], check=False))
            [[2, 2, 0, 0], [2, 1, 1, 0], [2, 1, 0, 1]]
        """
        return canonical_children(self._sgs, x, -1)

    def permutation_group(self):
        r"""
        Return the permutation group given to define ``self``.

        EXAMPLES::

            sage: I = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3,4)]]))
            sage: I.permutation_group()
            Permutation Group with generators [(1,2,3,4)]
        """
        return self._permgroup

    def is_canonical(self, v, check=True):
        r"""
        Return ``True`` if the integer list ``v`` is maximal in its
        orbit under the action of the permutation group given to
        define ``self``.  Such integer vectors are said to be
        canonical. A vector `v` is canonical if and only if

        .. MATH::

            v = \max_{\text{lex order}} \{g \cdot v | g \in G \}

        EXAMPLES::

            sage: I = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3,4)]]))
            sage: I.is_canonical([4,3,2,1])
            True
            sage: I.is_canonical([4,0,0,1])
            True
            sage: I.is_canonical([4,0,3,3])
            True
            sage: I.is_canonical([4,0,4,4])
            False
        """
        if check:
            assert isinstance(v, (ClonableIntArray, list)), '%s should be a list or an integer vector' % v
            assert (self.n == len(v)), '%s should be of length %s' % (v, self.n)
            for p in v:
                assert (p == NN(p)), 'Elements of %s should be integers' % v
        return is_canonical(self._sgs, self.element_class(self, list(v), check=False))

    def __contains__(self, v):
        """
        EXAMPLES::

            sage: I = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3,4)]]))
            sage: [2,2,0,0] in I
            True
            sage: [2,0,1,0] in I
            True
            sage: [2,0,0,1] in I
            True
            sage: [2,0,0,2] in I
            False
            sage: [2,0,0,2,12] in I
            False
        """
        try:
            return self.is_canonical(self.element_class(self, list(v), check=False), check=False)
        except Exception:
            return False

    def __call__(self, v, check=True):
        r"""
        Return an element of ``self`` constructed from ``v`` if
        possible.

        TESTS::

            sage: I = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3,4)]]))
            sage: I([3,2,1,0])
            [3, 2, 1, 0]
        """
        try:
            if v.parent() is self:
                return v
            else:
                raise ValueError('%s should be a Python list of integer' % (v))
        except Exception:
            return self.element_class(self, list(v), check=check)

    def orbit(self, v):
        r"""
        Return the orbit of the integer vector ``v`` under the action of the
        permutation group defining ``self``. The result is a set.

        EXAMPLES:

        In order to get reproducible doctests, we convert the returned sets
        into lists in increasing lexicographic order::

            sage: I = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3,4)]]))
            sage: sorted(I.orbit([2,2,0,0]))
            [[0, 0, 2, 2], [0, 2, 2, 0], [2, 0, 0, 2], [2, 2, 0, 0]]
            sage: sorted(I.orbit([2,1,0,0]))
            [[0, 0, 2, 1], [0, 2, 1, 0], [1, 0, 0, 2], [2, 1, 0, 0]]
            sage: sorted(I.orbit([2,0,1,0]))
            [[0, 1, 0, 2], [0, 2, 0, 1], [1, 0, 2, 0], [2, 0, 1, 0]]
            sage: sorted(I.orbit([2,0,2,0]))
            [[0, 2, 0, 2], [2, 0, 2, 0]]
            sage: I.orbit([1,1,1,1])
            {[1, 1, 1, 1]}
        """
        assert isinstance(v, (list, ClonableIntArray)), '%s should be a Python list or an element of %s' % (v, self)
        try:
            if v.parent() is self:
                return orbit(self._sgs, v)
            raise TypeError
        except Exception:
            return orbit(self._sgs, self.element_class(self, v, check=False))

    def subset(self, sum=None, max_part=None):
        r"""
        Return the subset of ``self`` containing integer vectors
        whose entries sum to ``sum``.

        EXAMPLES::

            sage: S = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3,4)]]))
            sage: S.subset(4)
            Integer vectors of length 4 and of sum 4 enumerated up to
            the action of Permutation Group with generators
            [(1,2,3,4)]
        """
        return IntegerVectorsModPermutationGroup_with_constraints(self.permutation_group(), sum, max_part)

    class Element(ClonableIntArray):
        r"""
        Element class for the set of integer vectors of given sum enumerated modulo
        the action of a permutation group. These vectors are clonable lists of integers
        which must satisfy conditions coming from the parent appearing in the method
        :meth:`~sage.structure.list_clone.ClonableIntArray.check`.

        TESTS::

            sage: I = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3,4)]]))
            sage: v = I.element_class(I, [4,3,2,1]); v
            [4, 3, 2, 1]
            sage: TestSuite(v).run()
            sage: I.element_class(I, [4,3,2,5])
            Traceback (most recent call last):
            ...
            AssertionError
        """

        def check(self):
            r"""
            Check that ``self`` verify the invariants needed for
            living in ``self.parent()``.

            EXAMPLES::

                sage: I = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3,4)]]))
                sage: v = I.an_element()
                sage: v.check()
                sage: w = I([0,4,0,0], check=False); w
                [0, 4, 0, 0]
                sage: w.check()
                Traceback (most recent call last):
                ...
                AssertionError
            """
            assert self.parent().is_canonical(self)


class IntegerVectorsModPermutationGroup_with_constraints(UniqueRepresentation, RecursivelyEnumeratedSet_forest):
    r"""
    This class models finite enumerated sets of integer vectors with
    constraint enumerated up to the action of a permutation group.
    Integer vectors are enumerated modulo the action of the
    permutation group. To implement that, we keep a single integer
    vector by orbit under the action of the permutation
    group. Elements chosen are vectors maximal in their orbit for the
    lexicographic order.

    For more information see :class:`IntegerVectorsModPermutationGroup`.

    EXAMPLES::

        sage: I = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3,4)]]),
        ....:                                       max_part=1)
        sage: I.list()
        [[0, 0, 0, 0], [1, 0, 0, 0], [1, 1, 0, 0], [1, 0, 1, 0], [1, 1, 1, 0],
         [1, 1, 1, 1]]
        sage: I = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3,4)]]),
        ....:                                       sum=6, max_part=4)
        sage: I.list()
        [[4, 2, 0, 0], [4, 1, 1, 0], [4, 1, 0, 1], [4, 0, 2, 0], [4, 0, 1, 1],
         [4, 0, 0, 2], [3, 3, 0, 0], [3, 2, 1, 0], [3, 2, 0, 1], [3, 1, 2, 0],
         [3, 1, 1, 1], [3, 1, 0, 2], [3, 0, 3, 0], [3, 0, 2, 1], [3, 0, 1, 2],
         [2, 2, 2, 0], [2, 2, 1, 1], [2, 1, 2, 1]]

    Here is the enumeration of unlabeled graphs over 5 vertices::

        sage: G = IntegerVectorsModPermutationGroup(TransitiveGroup(10,12), max_part=1)
        sage: G.cardinality()
        34

    TESTS::

        sage: I = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3,4)]]),4)
        sage: TestSuite(I).run()
    """

    def __init__(self, G, d, max_part, sgs=None):
        r"""
        TESTS::

            sage: I = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3,4)]]), 6, max_part=4)
        """
        RecursivelyEnumeratedSet_forest.__init__(self, algorithm='breadth', category=(FiniteEnumeratedSets(), FiniteEnumeratedSets().Quotients()))
        self._permgroup = G
        self.n = G.degree()
        self._sum = d
        if max_part is None:
            self._max_part = -1
        else:
            self._max_part = max_part

        # self.sgs: strong_generating_system
        if sgs is None:
            self._sgs = G.strong_generating_system()
        else:
            self._sgs = [list(x) for x in list(sgs)]

    def _repr_(self):
        r"""
        TESTS::

            sage: S = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3,4)]])); S
            Integer vectors of length 4 enumerated up to the action of Permutation Group with generators [(1,2,3,4)]
            sage: S = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3,4)]]), 6); S
            Integer vectors of length 4 and of sum 6 enumerated up to the action of Permutation Group with generators [(1,2,3,4)]
            sage: S = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3,4)]]), 6, max_part=4); S
            Vectors of length 4 and of sum 6 whose entries are in {0, ..., 4} enumerated up to the action of Permutation Group with generators [(1,2,3,4)]
            sage: S = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3,4)]]), max_part=4); S
            Integer vectors of length 4 whose entries are in {0, ..., 4} enumerated up to the action of Permutation Group with generators [(1,2,3,4)]
        """
        if self._sum is not None:
            if self._max_part >= 0:
                return ("Vectors of length %s and of sum %s"
                        " whose entries are in {0, ..., %s}"
                        " enumerated up to the action of %s"
                        % (self.n, self._sum, self._max_part,
                           self.permutation_group()))
            else:
                return ("Integer vectors of length %s"
                        " and of sum %s"
                        " enumerated up to the action of %s"
                        % (self.n, self._sum, self.permutation_group()))
        else:
            return ("Integer vectors of length %s"
                    " whose entries are in {0, ..., %s}"
                    " enumerated up to the action of %s"
                    % (self.n, self._max_part, self.permutation_group()))

    def roots(self):
        r"""
        Return the root of generation of ``self``.

        This method is
        required to build the tree structure of ``self`` which
        inherits from the class
        :class:`~sage.sets.recursively_enumerated_set.RecursivelyEnumeratedSet_forest`.

        EXAMPLES::

            sage: I = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3,4)]]))
            sage: I.roots()
            [[0, 0, 0, 0]]
        """
        return [self.element_class(self, self.n*[0,], check=False)]

    def children(self, x):
        r"""
        Return the list of children of the element ``x``.

        This method
        is required to build the tree structure of ``self`` which
        inherits from the class
        :class:`~sage.sets.recursively_enumerated_set.RecursivelyEnumeratedSet_forest`.

        EXAMPLES::

            sage: I = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3,4)]]))
            sage: I.children(I([2,1,0,0], check=False))
            [[2, 2, 0, 0], [2, 1, 1, 0], [2, 1, 0, 1]]
        """
        return canonical_children(self._sgs, x, -1)

    def permutation_group(self):
        r"""
        Return the permutation group given to define ``self``.

        EXAMPLES::

            sage: I = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3)]]), 5)
            sage: I.permutation_group()
            Permutation Group with generators [(1,2,3)]
        """
        return self._permgroup

    def __contains__(self, v):
        r"""
        TESTS::

            sage: I = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3,4)]]),6)
            sage: [6,0,0,0] in I
            True
            sage: [5,0,1,0] in I
            True
            sage: [0,5,1,0] in I
            False
            sage: [3,0,1,3] in I
            False
            sage: [3,3,1,0] in I
            False
        """
        try:
            return (self(v)).parent() is self
        except Exception:
            return False

    def __call__(self, v, check=True):
        r"""
        Make `v` an element living in ``self``.

        TESTS::

            sage: I = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3,4)]]), 4)
            sage: v = I([2,1,0,1]); v
            [2, 1, 0, 1]
            sage: v.parent()
            Integer vectors of length 4 and of sum 4 enumerated up to
            the action of Permutation Group with generators
            [(1,2,3,4)]
        """
        try:
            if v.parent() is self:
                return v
            else:
                raise ValueError('%s should be a Python list of integer' % (v))
        except Exception:
            return self.element_class(self, list(v), check=check)

    def __iter__(self):
        r"""
        TESTS::

            sage: I = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3,4)]]),4)
            sage: for i in I: i
            [4, 0, 0, 0]
            [3, 1, 0, 0]
            [3, 0, 1, 0]
            [3, 0, 0, 1]
            [2, 2, 0, 0]
            [2, 1, 1, 0]
            [2, 1, 0, 1]
            [2, 0, 2, 0]
            [2, 0, 1, 1]
            [1, 1, 1, 1]

            sage: I = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3,4)]]), sum=7, max_part=3)
            sage: for i in I: i
            [3, 3, 1, 0]
            [3, 3, 0, 1]
            [3, 2, 2, 0]
            [3, 2, 1, 1]
            [3, 2, 0, 2]
            [3, 1, 3, 0]
            [3, 1, 2, 1]
            [3, 1, 1, 2]
            [3, 0, 2, 2]
            [2, 2, 2, 1]

        Check that :issue:`36681` is fixed::

            sage: G = PermutationGroup([], domain=[])
            sage: I = IntegerVectorsModPermutationGroup(G, sum=0)
            sage: list(iter(I))
            [[]]

        Check that :issue:`36681` is fixed::

            sage: G = PermutationGroup([], domain=[])
            sage: I = IntegerVectorsModPermutationGroup(G, sum=3)
            sage: list(iter(I))
            []
        """
        # Special cases when domain is empty.
        if self.n == 0:
            if self._sum is not None and self._sum > 0:
                # No empty vector can have positive sum.
                return iter(())
            else:
                # Sum is allowed to be zero.  It does not matter what
                # the maxpart is, the empty vector is a solution.
                return iter([self([])])

        # General case, nonempty domain.
        if self._max_part < 0:
            return self.elements_of_depth_iterator(self._sum)
        else:
            SF = RecursivelyEnumeratedSet_forest(
                (self([0]*(self.n), check=False),),
                lambda x: [self(y, check=False)
                           for y in canonical_children(
                                   self._sgs, x, self._max_part)],
                algorithm='breadth')
            if self._sum is None:
                return iter(SF)
            else:
                return SF.elements_of_depth_iterator(self._sum)

    def cardinality(self):
        r"""
        Return the number of integer vectors in the set.

        The algorithm utilises :wikipedia:`Cycle Index Theorem <Cycle_index>`, allowing
        for a faster than a plain enumeration computation.

        EXAMPLES:

        With a trivial group all vectors are canonical::

            sage: G = PermutationGroup([], domain=[1,2,3])
            sage: IntegerVectorsModPermutationGroup(G, 5).cardinality()
            21
            sage: IntegerVectors(5, 3).cardinality()
            21

        With two interchangeable elements, the smaller one
        ranges from zero to ``sum//2``::

            sage: G = PermutationGroup([(1,2)])
            sage: IntegerVectorsModPermutationGroup(G, 1000).cardinality()
            501

        Binary vectors up to full symmetry are first some ones and
        then some zeros::

            sage: G = SymmetricGroup(10)
            sage: I = IntegerVectorsModPermutationGroup(G, max_part=1)
            sage: I.cardinality()
            11

        Binary vectors of constant weight, up to PGL(2,17), which
        is 3-transitive, but not 4-transitive::

            sage: G=PGL(2,17)
            sage: I = IntegerVectorsModPermutationGroup(G, sum=3, max_part=1)
            sage: I.cardinality()
            1
            sage: I = IntegerVectorsModPermutationGroup(G, sum=4, max_part=1)
            sage: I.cardinality()
            3

        TESTS:

        Check that :issue:`36681` is fixed::

            sage: G = PermutationGroup([], domain=[])
            sage: sgs = tuple(tuple(t) for t in G.strong_generating_system())
            sage: V = IntegerVectorsModPermutationGroup(G, sum=1, sgs=sgs)
            sage: V.cardinality()
            0

        The case when both ``sum`` and ``max_part`` are specified::

            sage: G = PermutationGroup([(1,2,3)])
            sage: I = IntegerVectorsModPermutationGroup(G, sum=10, max_part=5)
            sage: I.cardinality()
            7

        All permutation groups of degree 4::

            sage: for G in SymmetricGroup(4).subgroups():
            ....:     sgs = tuple(tuple(t) for t in G.strong_generating_system())
            ....:     I1 = IntegerVectorsModPermutationGroup(G, sum=10, sgs=sgs)
            ....:     assert I1.cardinality() == len(list(I1))
            ....:     I2 = IntegerVectorsModPermutationGroup(G, max_part=3, sgs=sgs)
            ....:     assert I2.cardinality() == len(list(I2))
            ....:     I3 = IntegerVectorsModPermutationGroup(G, sum=10, max_part=3, sgs=sgs)
            ....:     assert I3.cardinality() == len(list(I3))

        Symmetric group with sums 0 and 1::

            sage: S10 = SymmetricGroup(10)
            sage: IntegerVectorsModPermutationGroup(S10, 0).cardinality()
            1
            sage: IntegerVectorsModPermutationGroup(S10, 1).cardinality()
            1

        Trivial group with sums 1 and 100::

            sage: T10 = PermutationGroup([], domain=range(1, 11))
            sage: IntegerVectorsModPermutationGroup(T10, 1).cardinality()
            10
            sage: IntegerVectorsModPermutationGroup(T10, 100).cardinality()
            4263421511271
        """
        G = self._permgroup
        k = G.degree()          # Vector length
        d = self._sum           # Required sum
        m = self._max_part      # Max of one entry, -1 for no limit
        if m == -1:
            m = d               # Any entry cannot exceed total

        # Some easy special cases.
        if k == 0:
            # Empty vectors.  There is only one, and it has zero sum.
            # Here _max_part does not matter because any _max_part
            # condition is vacuously true (with no parts).
            if d == 0 or d is None:
                return Integer(1)
            else:
                return Integer(0)
        if d == 0 or m == 0:
            # All-zero vectors.  There is only one of them.
            return Integer(1)
        if d == 1:
            # Vectors with one 1 and all other elements zero.
            # The 1 can be placed in any orbit, and by symmetry
            # it will be on the first element of the orbit.
            return Integer(len(G.orbits()))
        if d is not None and m >= d and G.is_trivial():
            # Simple calculation with stars and bars.
            return Integer(binomial(d + k - 1, k - 1))

        # General case.
        #
        # Cardinality is computed using the Cycle Index Theorem.  We
        # have two cases.  With a fixed sum d we work with power
        # series and extract the x^d coefficient.  Without a fixed sum
        # we can do with integer arithmetic.
        Z = G.cycle_index()

        if d is None:
            # Case 1.  Without a fixed sum, the sum can be up to k*m.
            result = sum(coeff * (m+1)**len(cycle_type)
                         for cycle_type, coeff in Z)
            # Computed as Rational, but should have an integer value
            # by now.
            return Integer(result)

        # Case 2.  Fixed sum d.  Work with power series with enough
        # precision that x^d is valid.
        R = PowerSeriesRing(QQ, 'x', default_prec=d+1)
        x = R.gen()

        # The figure-counting series, for max_part==m, is (1-t**(m+1))
        # / (1-t) = 1+t+...+t**m.  For the function-counting series,
        # we substitute x**cycle_length for t.
        #
        funcount = sum(
            coeff * prod((1 - x**((m+1)*cycle_len)) / (1 - x**cycle_len)
                         for cycle_len in cycle_type)
            for cycle_type, coeff in Z)

        # Extract the d'th degree coefficient.  Computed as Rational,
        # but should have an integer value by now.
        return Integer(funcount[d])

    def is_canonical(self, v, check=True):
        r"""
        Return ``True`` if the integer list ``v`` is maximal in its
        orbit under the action of the permutation group given to
        define ``self``.  Such integer vectors are said to be
        canonical. A vector `v` is canonical if and only if

        .. MATH::

            v = \max_{\text{lex order}} \{g \cdot v | g \in G \}

        EXAMPLES::

            sage: I = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3,4)]]),
            ....:                                       max_part=3)
            sage: I.is_canonical([3,0,0,0])
            True
            sage: I.is_canonical([1,0,2,0])
            False
            sage: I.is_canonical([2,0,1,0])
            True
        """
        if check:
            assert isinstance(v, (ClonableIntArray, list)), '%s should be a list or an integer vector' % v
            assert (self.n == len(v)), '%s should be of length %s' % (v, self.n)
            for p in v:
                assert (p == NN(p)), 'Elements of %s should be integers' % v
        return is_canonical(self._sgs, self.element_class(self, list(v), check=False))

    def ambient(self):
        r"""
        Return the ambient space from which ``self`` is a quotient.

        EXAMPLES::

            sage: S = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3,4)]]), 6)
            sage: S.ambient()
            Integer vectors that sum to 6
            sage: S = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3,4)]]),
            ....:                                       6, max_part=12)
            sage: S.ambient()
            Integer vectors that sum to 6 with constraints: max_part=12
            sage: S = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3,4)]]),
            ....:                                       max_part=12)
            sage: S.ambient()
            Integer vectors with constraints: max_part=12
        """
        if self._sum is not None:
            if self._max_part <= -1:
                return IntegerVectors(n=self._sum)
            else:
                return IntegerVectors(n=self._sum, max_part=self._max_part)
        else:
            return IntegerVectors(max_part=self._max_part)

    def lift(self, elt):
        r"""
        Lift the element ``elt`` inside the ambient space from which ``self`` is a quotient.

        EXAMPLES::

            sage: S = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3,4)]]),
            ....:                                       max_part=1)
            sage: v = S.lift([1,0,1,0]); v
            [1, 0, 1, 0]
            sage: v in IntegerVectors(2,4,max_part=1)
            True
            sage: S = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3,4)]]),
            ....:                                       sum=6)
            sage: v = S.lift(S.list()[5]); v
            [4, 1, 1, 0]
            sage: v in IntegerVectors(n=6)
            True
        """
        # TODO: For now, Sage integer vectors are just python list.
        # Once Integer vectors will have an element class, update this
        # code properly
        return list(elt)

    def retract(self, elt):
        r"""
        Return the canonical representative of the orbit of the
        integer ``elt`` under the action of the permutation group
        defining ``self``.

        If the element ``elt`` is already maximal in its orbits for
        the lexicographic order, ``elt`` is thus the good
        representative for its orbit.

        EXAMPLES::

            sage: S = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3,4)]]),
            ....:                                       sum=2, max_part=1)
            sage: S.retract([1,1,0,0])
            [1, 1, 0, 0]
            sage: S.retract([1,0,1,0])
            [1, 0, 1, 0]
            sage: S.retract([1,0,0,1])
            [1, 1, 0, 0]
            sage: S.retract([0,1,1,0])
            [1, 1, 0, 0]
            sage: S.retract([0,1,0,1])
            [1, 0, 1, 0]
            sage: S.retract([0,0,1,1])
            [1, 1, 0, 0]
        """
        # TODO: Once Sage integer vector will have a data structure
        # based on ClonableIntArray, remove the conversion intarray
        assert len(elt) == self.n, "%s is a quotient set of %s" % (self, self.ambient())
        if self._sum is not None:
            assert sum(elt) == self._sum, "%s is a quotient set of %s" % (self, self.ambient())
        if self._max_part >= 0:
            assert max(elt) <= self._max_part, "%s is a quotient set of %s" % (self, self.ambient())
        intarray = self.element_class(self, elt, check=False)
        return self.element_class(self, canonical_representative_of_orbit_of(self._sgs, intarray), check=False)

    def _an_element_(self):
        r"""
        Return an element of ``self``.

        This raises an :exc:`EmptySetError` when ``self`` is empty.

        EXAMPLES::

            sage: S = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3,4)]]),
            ....:                                       sum=0, max_part=1)
            sage: S.an_element()
            [0, 0, 0, 0]
            sage: S = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3,4)]]),
            ....:                                       sum=1, max_part=1)
            sage: S.an_element()
            [1, 0, 0, 0]
            sage: S = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3,4)]]),
            ....:                                       sum=2, max_part=1)
            sage: S.an_element()
            [1, 1, 0, 0]
            sage: S = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3,4)]]),
            ....:                                       sum=3, max_part=1)
            sage: S.an_element()
            [1, 1, 1, 0]
            sage: S = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3,4)]]),
            ....:                                       sum=4, max_part=1)
            sage: S.an_element()
            [1, 1, 1, 1]
            sage: S = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3,4)]]),
            ....:                                       sum=5, max_part=1)
            sage: S.an_element()
            Traceback (most recent call last):
            ...
            EmptySetError
        """
        if self._max_part < 0:
            return self([self._sum]+(self.n-1)*[0], check=False)
        else:
            try:
                v = iter(self)
                return next(v)
            except StopIteration:
                from sage.categories.sets_cat import EmptySetError
                raise EmptySetError

    def orbit(self, v):
        r"""
        Return the orbit of the vector ``v`` under the action of the
        permutation group defining ``self``. The result is a set.

        INPUT:

        - ``v`` -- an element of ``self`` or any list of length the
          degree of the permutation group

        EXAMPLES:

        We convert the result in a list in increasing lexicographic
        order, to get a reproducible doctest::

            sage: I = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3,4)]]), 4)
            sage: I.orbit([1,1,1,1])
            {[1, 1, 1, 1]}
            sage: sorted(I.orbit([3,0,0,1]))
            [[0, 0, 1, 3], [0, 1, 3, 0], [1, 3, 0, 0], [3, 0, 0, 1]]
        """
        assert isinstance(v, (list, ClonableIntArray)), '%s should be a Python list or an element of %s' % (v, self)
        try:
            if v.parent() is self:
                return orbit(self._sgs, v)
        except Exception:
            return orbit(self._sgs, self.element_class(self, v, check=False))

    class Element(ClonableIntArray):
        r"""
        Element class for the set of integer vectors with constraints enumerated
        modulo the action of a permutation group. These vectors are clonable lists
        of integers which must satisfy conditions coming from the parent as in
        the method :meth:`~sage.combinat.integer_vectors_mod_permgroup.IntegerVectorsModPermutationGroup_with_constraints.Element.check`.

        TESTS::

            sage: I = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3,4)]]), 4)
            sage: v = I.element_class(I, [3,1,0,0]); v
            [3, 1, 0, 0]
            sage: TestSuite(v).run()
            sage: v = I.element_class(I, [3,2,0,0])
            Traceback (most recent call last):
            ...
            AssertionError: [3, 2, 0, 0] should be an integer vector of sum 4
        """

        def check(self):
            r"""
            Check that ``self`` meets the constraints of being an element of ``self.parent()``.

            EXAMPLES::

                sage: I = IntegerVectorsModPermutationGroup(PermutationGroup([[(1,2,3,4)]]), 4)
                sage: v = I.an_element()
                sage: v.check()
                sage: w = I([0,4,0,0], check=False); w
                [0, 4, 0, 0]
                sage: w.check()
                Traceback (most recent call last):
                ...
                AssertionError
            """
            if self.parent()._sum is not None:
                assert sum(self) == self.parent()._sum, '%s should be an integer vector of sum %s' % (self, self.parent()._sum)
            if self.parent()._max_part >= 0:
                assert max(self) <= self.parent()._max_part, 'Entries of %s must be inferior to %s' % (self, self.parent()._max_part)
            assert self.parent().is_canonical(self)



# sage_setup: distribution = sagemath-combinat
# sage.doctest: needs sage.combinat sage.graphs sage.modules
r"""
Bijection classes for type `A_{2n}^{(2)\dagger}`

Part of the (internal) classes which runs the bijection between rigged
configurations and KR tableaux of type `A_{2n}^{(2)\dagger}`.

AUTHORS:

- Travis Scrimshaw (2012-12-21): Initial version

TESTS::

    sage: KRT = crystals.TensorProductOfKirillovReshetikhinTableaux(CartanType(['A', 4, 2]).dual(), [[2, 1]])
    sage: from sage.combinat.rigged_configurations.bij_type_A2_dual import KRTToRCBijectionTypeA2Dual
    sage: bijection = KRTToRCBijectionTypeA2Dual(KRT(pathlist=[[2,1]]))
    sage: TestSuite(bijection).run()
    sage: RC = RiggedConfigurations(CartanType(['A', 4, 2]).dual(), [[2, 1]])
    sage: from sage.combinat.rigged_configurations.bij_type_A2_dual import RCToKRTBijectionTypeA2Dual
    sage: bijection = RCToKRTBijectionTypeA2Dual(RC(partition_list=[[],[]]))
    sage: TestSuite(bijection).run()
"""

# ****************************************************************************
#       Copyright (C) 2012 Travis Scrimshaw <tscrim@ucdavis.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.combinat.rigged_configurations.bij_type_C import KRTToRCBijectionTypeC
from sage.combinat.rigged_configurations.bij_type_C import RCToKRTBijectionTypeC
from sage.combinat.rigged_configurations.bij_type_A import KRTToRCBijectionTypeA

from sage.rings.rational_field import QQ


class KRTToRCBijectionTypeA2Dual(KRTToRCBijectionTypeC):
    r"""
    Specific implementation of the bijection from KR tableaux to rigged
    configurations for type `A_{2n}^{(2)\dagger}`.

    This inherits from type `C_n^{(1)}` because we use the same methods in
    some places.
    """

    def next_state(self, val):
        r"""
        Build the next state for type `A_{2n}^{(2)\dagger}`.

        TESTS::

            sage: KRT = crystals.TensorProductOfKirillovReshetikhinTableaux(CartanType(['A', 4, 2]).dual(), [[2,1]])
            sage: from sage.combinat.rigged_configurations.bij_type_A2_dual import KRTToRCBijectionTypeA2Dual
            sage: bijection = KRTToRCBijectionTypeA2Dual(KRT(pathlist=[[-1,2]]))
            sage: bijection.cur_path.insert(0, [])
            sage: bijection.cur_dims.insert(0, [0, 1])
            sage: bijection.cur_path[0].insert(0, [2])
            sage: bijection.next_state(2)
        """
        n = self.n
        tableau_height = len(self.cur_path[0]) - 1

        if val > 0:
            # If it is a regular value, we follow the A_n rules
            KRTToRCBijectionTypeA.next_state(self, val)
            return

        pos_val = -val

        if pos_val == 0:
            if len(self.ret_rig_con[pos_val - 1]) > 0:
                max_width = self.ret_rig_con[n-1][0]
            else:
                max_width = 1
            max_width = self.ret_rig_con[n-1].insert_cell(max_width)
            width_n = max_width + 1

            # Follow regular A_n rules
            for a in reversed(range(tableau_height, n-1)):
                max_width = self.ret_rig_con[a].insert_cell(max_width)
                self._update_vacancy_nums(a + 1)
                self._update_partition_values(a + 1)
            self._update_vacancy_nums(tableau_height)
            self._update_partition_values(tableau_height)
            if tableau_height > 0:
                self._update_vacancy_nums(tableau_height-1)
                self._update_partition_values(tableau_height-1)

            # Make the new string at n quasi-singular
            p = self.ret_rig_con[n-1]
            for i in range(len(p)):
                if p._list[i] == width_n:
                    p.rigging[i] = p.rigging[i] - QQ(1)/QQ(2)
                    break
            return

        case_S = [None] * n
        pos_val = -val

        # Always add a cell to the first singular value in the first
        #   tableau we are updating.
        if len(self.ret_rig_con[pos_val - 1]) > 0:
            max_width = self.ret_rig_con[pos_val - 1][0]
        else:
            max_width = 1

        # Add cells similar to type A_n but we move to the right until we
        #   reach the value of n-1
        for a in range(pos_val - 1, n-1):
            max_width = self.ret_rig_con[a].insert_cell(max_width)
            case_S[a] = max_width

        # Special case for n
        # If we find a quasi-singular string first, then we are in case (Q, S)
        #   otherwise we will find a singular string and insert 2 cells
        partition = self.ret_rig_con[n-1]
        num_rows = len(partition)
        case_QS = False
        for i in range(num_rows + 1):
            if i == num_rows:
                max_width = 0
                if case_QS:
                    partition._list.append(1)
                    partition.vacancy_numbers.append(None)
                    # Go through our partition until we find a length of greater than 1
                    j = len(partition._list) - 1
                    while j >= 0 and partition._list[j] == 1:
                        j -= 1
                    partition.rigging.insert(j + 1, None)
                    width_n = 1
                else:
                    # Go through our partition until we find a length of greater than 2
                    j = len(partition._list) - 1
                    while j >= 0 and partition._list[j] <= 2:
                        j -= 1
                    partition._list.insert(j+1, 2)
                    partition.vacancy_numbers.insert(j+1, None)
                    partition.rigging.insert(j+1, None)
                break
            elif partition._list[i] <= max_width:
                if partition.vacancy_numbers[i] == partition.rigging[i]:
                    max_width = partition._list[i]
                    if case_QS:
                        partition._list[i] += 1
                        width_n = partition._list[i]
                        partition.rigging[i] = None
                    else:
                        j = i - 1
                        while j >= 0 and partition._list[j] <= max_width + 2:
                            partition.rigging[j+1] = partition.rigging[j]  # Shuffle it along
                            j -= 1
                        partition._list.pop(i)
                        partition._list.insert(j+1, max_width + 2)
                        partition.rigging[j+1] = None
                    break
                elif partition.vacancy_numbers[i] - QQ(1)/QQ(2) == partition.rigging[i] and not case_QS:
                    case_QS = True
                    partition._list[i] += 1
                    partition.rigging[i] = None
                    # No need to set max_width here since we will find a singular string

        # Now go back following the regular C_n (ish) rules
        for a in reversed(range(tableau_height, n-1)):
            if case_S[a] == max_width:
                self._insert_cell_case_S(self.ret_rig_con[a])
            else:
                max_width = self.ret_rig_con[a].insert_cell(max_width)
            self._update_vacancy_nums(a + 1)
            self._update_partition_values(a + 1)

        # Update the final rigged partitions
        if tableau_height < n:
            self._update_vacancy_nums(tableau_height)
            self._update_partition_values(tableau_height)

        if pos_val <= tableau_height:
            for a in range(pos_val-1, tableau_height):
                self._update_vacancy_nums(a)
                self._update_partition_values(a)
            if pos_val > 1:
                self._update_vacancy_nums(pos_val - 2)
                self._update_partition_values(pos_val - 2)
        elif tableau_height > 0:
            self._update_vacancy_nums(tableau_height - 1)
            self._update_partition_values(tableau_height - 1)

        if case_QS:
            # Make the new string quasi-singular
            num_rows = len(partition)
            for i in range(num_rows):
                if partition._list[i] == width_n:
                    partition.rigging[i] = partition.rigging[i] - QQ(1)/QQ(2)
                    break


class RCToKRTBijectionTypeA2Dual(RCToKRTBijectionTypeC):
    r"""
    Specific implementation of the bijection from rigged configurations to
    tensor products of KR tableaux for type `A_{2n}^{(2)\dagger}`.
    """

    def next_state(self, height):
        r"""
        Build the next state for type `A_{2n}^{(2)\dagger}`.

        TESTS::

            sage: RC = RiggedConfigurations(CartanType(['A', 4, 2]).dual(), [[2,1]])
            sage: from sage.combinat.rigged_configurations.bij_type_A2_dual import RCToKRTBijectionTypeA2Dual
            sage: bijection = RCToKRTBijectionTypeA2Dual(RC(partition_list=[[2],[2,2]]))
            sage: bijection.next_state(2)
            -1
        """
        height -= 1  # indexing
        n = self.n
        ell = [None] * (2*n)
        case_S = [False] * n
        case_Q = False
        b = None

        # Calculate the rank and ell values

        last_size = 0
        for a in range(height, n-1):
            ell[a] = self._find_singular_string(self.cur_partitions[a], last_size)

            if ell[a] is None:
                b = a + 1
                break
            else:
                last_size = self.cur_partitions[a][ell[a]]

        if b is None:
            partition = self.cur_partitions[n-1]
            # Special case for n
            for i in reversed(range(len(partition))):
                if partition[i] >= last_size:
                    if partition.vacancy_numbers[i] == partition.rigging[i]:
                        last_size = partition[i]
                        case_S[n-1] = True
                        ell[2*n-1] = i
                        break
                    elif partition.vacancy_numbers[i] - QQ(1)/QQ(2) == partition.rigging[i] and not case_Q:
                        case_Q = True
                        # This will never be singular
                        last_size = partition[i] + 1
                        ell[n-1] = i

            if ell[2*n-1] is None:
                if not case_Q:
                    b = n
                else:
                    b = 0

        if b is None:
            # Now go back
            for a in reversed(range(n-1)):
                if a >= height and self.cur_partitions[a][ell[a]] == last_size:
                    ell[n+a] = ell[a]
                    case_S[a] = True
                else:
                    ell[n+a] = self._find_singular_string(self.cur_partitions[a], last_size)

                    if ell[n + a] is None:
                        b = -(a + 2)
                        break
                    else:
                        last_size = self.cur_partitions[a][ell[n + a]]

        if b is None:
            b = -1

        # Determine the new rigged configuration by removing boxes from the
        #   selected string and then making the new string singular
        if n > 1:
            if case_S[0]:
                row_num = None
                row_num_bar = self.cur_partitions[0].remove_cell(ell[n], 2)
            else:
                row_num = self.cur_partitions[0].remove_cell(ell[0])
                row_num_bar = self.cur_partitions[0].remove_cell(ell[n])
        for a in range(1, n-1):
            if case_S[a]:
                row_num_next = None
                row_num_bar_next = self.cur_partitions[a].remove_cell(ell[n+a], 2)
            else:
                row_num_next = self.cur_partitions[a].remove_cell(ell[a])
                row_num_bar_next = self.cur_partitions[a].remove_cell(ell[n+a])

            self._update_vacancy_numbers(a - 1)
            if row_num is not None:
                self.cur_partitions[a-1].rigging[row_num] = self.cur_partitions[a-1].vacancy_numbers[row_num]
            if row_num_bar is not None:
                self.cur_partitions[a-1].rigging[row_num_bar] = self.cur_partitions[a-1].vacancy_numbers[row_num_bar]
            row_num = row_num_next
            row_num_bar = row_num_bar_next

        if case_Q:
            row_num_next = self.cur_partitions[n-1].remove_cell(ell[n-1])
            if case_S[n-1]:
                row_num_bar_next = self.cur_partitions[n-1].remove_cell(ell[2*n-1])
            else:
                row_num_bar_next = None
        elif case_S[n-1]:
            row_num_next = None
            row_num_bar_next = self.cur_partitions[n-1].remove_cell(ell[2*n-1], 2)
        else:
            row_num_next = None
            row_num_bar_next = None

        if n > 1:
            self._update_vacancy_numbers(n - 2)
            if row_num is not None:
                self.cur_partitions[n-2].rigging[row_num] = self.cur_partitions[n-2].vacancy_numbers[row_num]
            if row_num_bar is not None:
                self.cur_partitions[n-2].rigging[row_num_bar] = self.cur_partitions[n-2].vacancy_numbers[row_num_bar]

        self._update_vacancy_numbers(n - 1)
        if row_num_next is not None:
            self.cur_partitions[n-1].rigging[row_num_next] = self.cur_partitions[n-1].vacancy_numbers[row_num_next]
        if row_num_bar_next is not None:
            if case_Q:
                # This will always be the largest value
                self.cur_partitions[n-1].rigging[row_num_bar_next] = self.cur_partitions[n-1].vacancy_numbers[row_num_bar_next] - QQ(1)/QQ(2)
            else:
                self.cur_partitions[n-1].rigging[row_num_bar_next] = self.cur_partitions[n-1].vacancy_numbers[row_num_bar_next]

        return b



# sage_setup: distribution = sagemath-combinat
# sage.doctest: needs sage.symbolic
r"""
Hexads in S(5, 6, 12)

This module completes a 5-element subset of a 12-set `X`
into a hexad in a Steiner system `S(5, 6, 12)` using Curtis
and Conway's "kitten method".  The labeling is either the
"modulo 11" labeling or the "shuffle" labeling.

The  main functions implemented in this file are
:meth:`Minimog.blackjack_move` and :meth:`Minimog.find_hexad`.

Enter ``blackjack_move?``
for help to play blackjack (i.e., the rules of the game), or
``find_hexad?`` for help finding hexads of `S(5, 6, 12)` in
the shuffle labeling.

This picture is the kitten in the "shuffle" labeling::


                                     6

                                     9

                                10     8

                             7     2      5

                         9     4     11     9

                      10     8     3      10     8

                 1                                      0


The corresponding MINIMOG is::

             ┌─────┬─────┬─────┬─────┐
             │  6  │  3  │  0  │  9  │
             ├─────┼─────┼─────┼─────┤
             │  5  │  2  │  7  │ 10  │
             ├─────┼─────┼─────┼─────┤
             │  4  │  1  │  8  │ 11  │
             └─────┴─────┴─────┴─────┘

which is specified by the global variable ``minimog_shuffle``.

See the docstrings for :meth:`Minimog.find_hexad` and
:meth:`Minimog.blackjack_move` for further details and examples.

AUTHOR:

David Joyner (2006-05)

REFERENCES:

- [Cu1984]_

- [Co1984]_

- [CS1986]_

- [KR2001]_

Some details are also online at:  https://www.permutationpuzzles.org/hexad/
"""
# ****************************************************************************
#       Copyright (C) 2005 David Joyner <wdjoyner@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL),
#  version 2 or later (at your choice)
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.rings.infinity import infinity
from sage.matrix.matrix_space import MatrixSpace
from sage.matrix.constructor import matrix
from sage.rings.finite_rings.finite_field_constructor import GF
from sage.calculus.calculus import SR


def view_list(L):
    """
    This provides a printout of the lines, crosses and squares
    of the MINIMOG, as in Curtis' paper [Cu1984]_.

    EXAMPLES::

        sage: from sage.games.hexad import *
        sage: M = Minimog(type='shuffle')
        sage: view_list(M.line[1])
        <BLANKLINE>
        [0 0 0]
        [1 1 1]
        [0 0 0]
        sage: view_list(M.cross[1])
        <BLANKLINE>
        [1 1 1]
        [0 0 1]
        [0 0 1]
        sage: view_list(M.square[1])
        <BLANKLINE>
        [0 0 0]
        [1 1 0]
        [1 1 0]
    """
    return matrix(GF(2), 3, 3, lambda x, y: 1 if (x, y) in L else 0)


def picture_set(A, L) -> set:
    """
    This is needed in the :meth:`Minimog.find_hexad` function below.

    EXAMPLES::

        sage: from sage.games.hexad import *
        sage: M = Minimog(type='shuffle')
        sage: picture_set(M.picture00, M.cross[2])
        {5, 7, 8, 9, 10}
        sage: picture_set(M.picture02, M.square[7])
        {2, 3, 5, 8}
    """
    return {A[x] for x in L}


class Minimog:
    r"""
    This implements the Conway/Curtis minimog idea for describing
    the Steiner triple system `S(5, 6, 12)`.

    EXAMPLES::

        sage: from sage.games.hexad import *
        sage: Minimog(type='shuffle')
        Minimog of type shuffle
        sage: M = Minimog(type = "modulo11")
        sage: M.minimog
        [        0         3 +Infinity         2]
        [        5         9         8        10]
        [        4         1         6         7]
    """
    def __init__(self, type='shuffle'):
        self.type = type
        MS34 = MatrixSpace(SR, 3, 4)
        minimog_modulo11 = MS34([[0, 3, infinity, 2], [5, 9, 8, 10], [4, 1, 6, 7]])
        minimog_shuffle = MS34([[6, 3, 0, 9], [5, 2, 7, 10], [4, 1, 8, 11]])
        if type == "shuffle":
            self.minimog = minimog_shuffle
        elif type == "modulo11":
            self.minimog = minimog_modulo11
        else:
            raise ValueError("that Minimog type is not implemented")
        # This initializes the variables in the game.
        MS34 = MatrixSpace(SR, 3, 4)
        A = self.minimog
        MS33 = MatrixSpace(SR, 3, 3)
        self.picture00 = MS33([[A[(1, 0)], A[(2, 3)], A[(0, 1)]], [A[(2, 2)], A[(1, 1)], A[(2, 0)]], [A[(0, 3)], A[(1, 3)], A[(1, 2)]]])
        # self.picture00 is the "picture at 6"
        self.picture02 = MS33([[A[(1, 0)], A[(2, 3)], A[(0, 1)]], [A[(1, 1)], A[(2, 0)], A[(2, 2)]], [A[(1, 2)], A[(0, 3)], A[(1, 3)]]])
        # self.picture02 is the "picture at 1"
        self.picture21 = MS33([[A[(2, 2)], A[(1, 3)], A[(0, 1)]], [A[(0, 3)], A[(2, 3)], A[(2, 0)]], [A[(1, 0)], A[(1, 1)], A[(1, 2)]]])
        # self.picture21 is the "picture at 0"

        self.line = list(range(12))
        self.line[0] = {(0, 0), (0, 1), (0, 2)}
        self.line[1] = {(1, 0), (1, 1), (1, 2)}
        self.line[2] = {(2, 0), (2, 1), (2, 2)}
        self.line[3] = {(0, 2), (1, 2), (2, 2)}
        self.line[4] = {(0, 1), (1, 1), (2, 1)}
        self.line[5] = {(0, 0), (1, 0), (2, 0)}
        self.line[6] = {(0, 0), (1, 1), (2, 2)}
        self.line[7] = {(2, 0), (0, 1), (1, 2)}
        self.line[8] = {(0, 2), (1, 0), (2, 1)}
        self.line[9] = {(2, 0), (1, 1), (0, 2)}
        self.line[10] = {(0, 0), (1, 2), (2, 1)}
        self.line[11] = {(1, 0), (0, 1), (2, 2)}

        self.cross = list(range(18))
        self.cross[0] = {(0, 0), (0, 1), (0, 2), (1, 0), (2, 0)}
        self.cross[1] = {(0, 0), (0, 1), (0, 2), (1, 2), (2, 2)}
        self.cross[2] = {(0, 0), (1, 0), (2, 0), (2, 1), (2, 2)}
        self.cross[3] = {(2, 0), (2, 1), (2, 2), (0, 2), (1, 2)}
        self.cross[4] = {(0, 0), (0, 1), (0, 2), (1, 1), (2, 1)}
        self.cross[5] = {(0, 0), (1, 0), (2, 0), (1, 1), (1, 2)}
        self.cross[6] = {(1, 0), (1, 1), (1, 2), (0, 2), (2, 2)}
        self.cross[7] = {(0, 1), (1, 1), (2, 1), (2, 0), (2, 2)}
        self.cross[8] = {(0, 0), (0, 1), (1, 0), (1, 1), (2, 2)}
        self.cross[9] = {(0, 0), (1, 1), (1, 2), (2, 1), (2, 2)}
        self.cross[10] = {(2, 0), (2, 1), (1, 0), (1, 1), (0, 2)}
        self.cross[11] = {(0, 1), (0, 2), (1, 1), (1, 2), (2, 0)}
        self.cross[12] = {(0, 0), (1, 0), (0, 2), (1, 2), (2, 1)}
        self.cross[13] = {(1, 0), (0, 1), (0, 2), (2, 1), (2, 2)}
        self.cross[14] = {(0, 1), (1, 0), (1, 2), (2, 0), (2, 2)}
        self.cross[15] = {(0, 0), (0, 1), (1, 2), (2, 0), (2, 1)}
        self.cross[16] = {(1, 0), (1, 1), (1, 2), (0, 1), (2, 1)}
        self.cross[17] = {(0, 0), (0, 2), (1, 1), (2, 0), (2, 2)}
        self.box = {(i, j) for i in range(3) for j in range(3)}
        self.square = [set() for i in range(18)]
        for i in range(18):
            self.square[i] = self.box - self.cross[i]

        MS34_GF3 = MatrixSpace(GF(2), 3, 4)
        cols = {}
        cols[1] = MS34_GF3([[1, 0, 0, 0], [1, 0, 0, 0], [1, 0, 0, 0]])
        cols[2] = MS34_GF3([[0, 1, 0, 0], [0, 1, 0, 0], [0, 1, 0, 0]])
        cols[3] = MS34_GF3([[0, 0, 1, 0], [0, 0, 1, 0], [0, 0, 1, 0]])
        cols[4] = MS34_GF3([[0, 0, 0, 1], [0, 0, 0, 1], [0, 0, 0, 1]])
        self.col = cols

        tets = {}
        tets[1] = MS34_GF3([[1, 1, 1, 1], [0, 0, 0, 0], [0, 0, 0, 0]])
        tets[2] = MS34_GF3([[1, 0, 0, 0], [0, 1, 1, 1], [0, 0, 0, 0]])
        tets[3] = MS34_GF3([[1, 0, 0, 0], [0, 0, 0, 0], [0, 1, 1, 1]])
        tets[4] = MS34_GF3([[0, 1, 0, 0], [1, 0, 1, 0], [0, 0, 0, 1]])
        tets[5] = MS34_GF3([[0, 0, 0, 1], [1, 1, 0, 0], [0, 0, 1, 0]])
        tets[6] = MS34_GF3([[0, 0, 1, 0], [1, 0, 0, 1], [0, 1, 0, 0]])
        tets[7] = MS34_GF3([[0, 1, 0, 0], [0, 0, 0, 1], [1, 0, 1, 0]])
        tets[8] = MS34_GF3([[0, 0, 1, 0], [0, 1, 0, 0], [1, 0, 0, 1]])
        tets[9] = MS34_GF3([[0, 0, 0, 1], [0, 0, 1, 0], [1, 1, 0, 0]])
        self.tet = tets

    def __repr__(self) -> str:
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: M = Minimog(type='modulo11')
            sage: M
            Minimog of type modulo11
        """
        return "Minimog of type %s" % self.type

    def __str__(self) -> str:
        """
        EXAMPLES::

            sage: M = Minimog(type='modulo11')
            sage: print(M)
            Minimog of type modulo11 associated to
            [        0         3 +Infinity         2]
            [        5         9         8        10]
            [        4         1         6         7]
        """
        return f"Minimog of type {self.type} associated to\n {self.minimog}"

    def _latex_(self):
        r"""
        Print latex code.

        EXAMPLES::

            sage: M = Minimog(type='modulo11')
            sage: latex(M)
            Minimog of type modulo11 associated to
             $\left(\begin{array}{rrrr}
            0 & 3 & +\infty & 2 \\
            5 & 9 & 8 & 10 \\
            4 & 1 & 6 & 7
            \end{array}\right)$
        """
        from sage.misc.latex import latex
        return f"Minimog of type {self.type} associated to\n ${latex(self.minimog)}$"

    def print_kitten(self):
        """
        This simply prints the "kitten" (expressed as a triangular diagram
        of symbols).

        EXAMPLES::

            sage: from sage.games.hexad import *
            sage: M = Minimog("shuffle")
            sage: M.print_kitten()
                     0
            <BLANKLINE>
                     8
                   9  10
                  5  11  3
                 8  2  4  8
               9  10  7  9  10
            <BLANKLINE>
            6                   1
            sage: M = Minimog("modulo11")
            sage: M.print_kitten()
                     +Infinity
            <BLANKLINE>
                     6
                   2  10
                  5  7  3
                 6  9  4  6
               2  10  8  2  10
            <BLANKLINE>
            0                   1
        """
        MINIMOG = self.minimog
        kitten = f'         {MINIMOG[0][2]}'
        kitten += '\n           '
        kitten += f'\n         {MINIMOG[2][2]}'
        kitten += f'\n       {MINIMOG[0][3]}  {MINIMOG[1][3]}'
        kitten += f'\n      {MINIMOG[1][0]}  {MINIMOG[2][3]}  {MINIMOG[0][1]}'
        kitten += '\n     {0}  {1}  {2}  {0}'.format(MINIMOG[2][2], MINIMOG[1][1], MINIMOG[2][0])
        kitten += '\n   {0}  {1}  {2}  {0}  {1}'.format(MINIMOG[0][3], MINIMOG[1][3], MINIMOG[1][2])
        kitten += '\n           \n'
        kitten += f'{MINIMOG[0][0]}                   {MINIMOG[2][1]}'
        print(kitten)

    def find_hexad0(self, pts):
        """
        Find a hexad of type 0.

        INPUT:

        - ``pts`` -- set of 2 distinct elements of MINIMOG, but not
          including the "points at infinity"

        OUTPUT:

        hexad containing ``pts`` and of type 0 (the 3 points "at
        infinity" union a line)

        .. NOTE::

            The 3 points "at infinity" are
            ``{MINIMOG[0][2], MINIMOG[2][1], MINIMOG[0][0]}``.

        EXAMPLES::

            sage: from sage.games.hexad import *
            sage: M = Minimog(type='shuffle')
            sage: M.find_hexad0(set([2, 4]))
            ([0, 1, 2, 4, 6, 8], ['line 1', 'picture 1'])
        """
        MINIMOG = self.minimog
        L = set(pts)
        H = {MINIMOG[0][2], MINIMOG[2][1], MINIMOG[0][0]}
        for i in range(12):
            if L <= picture_set(self.picture02, self.line[i]):
                WHAT = ["line " + str(i), "picture " + str(1)]
                H = H | picture_set(self.picture02, self.line[i])
                return list(H), WHAT
            if L <= picture_set(self.picture21, self.line[i]):
                WHAT = ["line " + str(i), "picture " + str(0)]
                H = H | picture_set(self.picture21, self.line[i])
                return list(H), WHAT
            if L <= picture_set(self.picture00, self.line[i]):
                WHAT = ["line " + str(i), "picture " + str(6)]
                H = H | picture_set(self.picture00, self.line[i])
                return list(H), WHAT
        return [], []

    def find_hexad1(self, pts):
        """
        Find a hexad of type 1.

        INPUT:

        - ``pts`` -- set of 5 distinct elements of MINIMOG

        OUTPUT:

        hexad containing ``pts`` and of type 1 (union of 2 parallel
        lines -- *no* points "at infinity")

        .. NOTE::

            The 3 points "at infinity" are
            ``{MINIMOG[0][2], MINIMOG[2][1], MINIMOG[0][0]}``.

        EXAMPLES::

            sage: from sage.games.hexad import *
            sage: M = Minimog(type='shuffle')
            sage: M.find_hexad1(set([2, 3, 4, 5, 8]))
            ([2, 3, 4, 5, 8, 11], ['lines (1, 2)', 'picture 1'])
        """
        H = set(pts)
        L = set(pts)
        linez = [(1, 2), (1, 3), (2, 3), (4, 5), (4, 6), (5, 6),
                 (7, 8), (7, 9), (8, 9), (10, 11), (10, 12), (11, 12)]
        for x in linez:
            x1 = int(x[0] - 1)
            x2 = int(x[1] - 1)      # (recall | is union)
            if L <= (picture_set(self.picture02, self.line[x1]) | picture_set(self.picture02, self.line[x2])):
                WHAT = ["lines " + str(x), "picture " + str(1)]
                H = picture_set(self.picture02, self.line[x1]) | picture_set(self.picture02, self.line[x2])
                return list(H), WHAT
            if L <= (picture_set(self.picture21, self.line[x1]) | picture_set(self.picture21, self.line[x2])):
                WHAT = ["lines " + str(x), "picture " + str(0)]
                H = picture_set(self.picture21, self.line[x1]) | picture_set(self.picture21, self.line[x2])
                return list(H), WHAT
            if L <= (picture_set(self.picture00, self.line[x1]) | picture_set(self.picture00, self.line[x2])):
                WHAT = ["lines " + str(x), "picture " + str(6)]
                H = picture_set(self.picture00, self.line[x1]) | picture_set(self.picture00, self.line[x2])
                return list(H), WHAT
        return [], []

    def find_hexad2(self, pts, x0):
        r"""
        Find a hexad of type 2.

        INPUT:

        - ``pts`` -- list S of 4 elements of MINIMOG, not including
          any "points at infinity"
        - ``x0`` -- in ``{MINIMOG[0][2], MINIMOG[2][1], MINIMOG[0][0]}``

        OUTPUT: hexad containing `S \cup \{x0\}` of type 2

        EXAMPLES::

            sage: from sage.games.hexad import *
            sage: M = Minimog(type='shuffle')
            sage: M.find_hexad2([2, 3, 4, 5], 1)
            ([], [])

        The above output indicates that there is no hexad of type 2
        containing `\{2, 3, 4, 5\}`. However, there is one containing
        `\{2, 3, 4, 8\}`::

            sage: M.find_hexad2([2, 3, 4, 8], 0)
            ([0, 2, 3, 4, 8, 9], ['cross 12', 'picture 0'])
        """
        MINIMOG = self.minimog
        L = set(pts)
        H = {x0}
        for i in range(18):
            if (x0 == MINIMOG[2][1] and L <= picture_set(self.picture02, self.cross[i])):
                WHAT = ["cross " + str(i), "picture " + str(1)]
                H = H | picture_set(self.picture02, self.cross[i])
                return list(H), WHAT
            if (x0 == MINIMOG[0][2] and L <= picture_set(self.picture21, self.cross[i])):
                WHAT = ["cross " + str(i), "picture " + str(MINIMOG[0][2])]
                H = H | picture_set(self.picture21, self.cross[i])
                return list(H), WHAT
            if (x0 == MINIMOG[0][0] and L <= picture_set(self.picture00, self.cross[i])):
                WHAT = ["cross " + str(i), "picture " + str(6)]
                H = H | picture_set(self.picture00, self.cross[i])
                return list(H), WHAT
        return [], []

    def find_hexad3(self, pts, x0, x1):
        r"""
        Find a hexad of type 3.

        INPUT:

        - ``pts`` -- list of 3 elements of MINIMOG, not including any
          "points at infinity"
        - ``x0``, ``x1`` -- in ``{MINIMOG[0][2], MINIMOG[2][1],
          MINIMOG[0][0]}``

        OUTPUT:

        hexad containing pts union \{x0, x1\} of type 3 (square at
        picture of "omitted point at infinity")

        EXAMPLES::

            sage: from sage.games.hexad import *
            sage: M = Minimog(type='shuffle')
            sage: M.find_hexad3([2, 3, 4], 0, 1)
            ([0, 1, 2, 3, 4, 11], ['square 2', 'picture 6'])
        """
        MINIMOG = self.minimog
        L = set(pts)
        H = {x0, x1}
        for i in range(18):
            if (MINIMOG[0][2] not in H and L <= picture_set(self.picture21, self.square[i])):
                WHAT = ["square " + str(i), "picture " + str(MINIMOG[0][2])]
                H = H | picture_set(self.picture21, self.square[i])
                return list(H), WHAT
            if (MINIMOG[2][1] not in H and L <= picture_set(self.picture02, self.square[i])):
                WHAT = ["square " + str(i), "picture " + str(MINIMOG[2][1])]
                H = H | picture_set(self.picture02, self.square[i])
                return list(H), WHAT
            if (MINIMOG[0][0] not in H and L <= picture_set(self.picture00, self.square[i])):
                WHAT = ["square " + str(i), "picture " + str(MINIMOG[0][0])]
                H = H | picture_set(self.picture00, self.square[i])
                return list(H), WHAT
        return [], []

    def find_hexad(self, pts):
        r"""
        Find a hexad of some type.

        INPUT:

        - ``pts`` -- list S of 5 elements of MINIMOG

        OUTPUT: hexad containing `S \cup \{x0\}` of some type

        .. NOTE::

            The 3 "points at infinity" are
            ``{MINIMOG[0][2], MINIMOG[2][1], MINIMOG[0][0]}``.

        Theorem ([Cu1984]_,  [Co1984]_): Each hexads is of exactly one
        of the following types:

        0. {3 "points at infinity"} union {any line},
        1. the union of any two (distinct) parallel lines in the same
           picture,
        2. one "point at infinity" union a cross in the corresponding
           picture, or
        3. two "points at infinity" union a square in the picture
           corresponding to the omitted point at infinity.

        More precisely, there are 132 such hexads (12 of type 0,
        12 of type 1, 54 of type 2, and 54 of type 3).
        They form a Steiner system of type `(5, 6, 12)`.

        EXAMPLES::

            sage: from sage.games.hexad import *
            sage: M = Minimog(type='shuffle')
            sage: M.find_hexad([0, 1, 2, 3, 4])
            ([0, 1, 2, 3, 4, 11], ['square 2', 'picture 6'])
            sage: M.find_hexad([1, 2, 3, 4, 5])
            ([1, 2, 3, 4, 5, 6], ['square 8', 'picture 0'])
            sage: M.find_hexad([2, 3, 4, 5, 8])
            ([2, 3, 4, 5, 8, 11], ['lines (1, 2)', 'picture 1'])
            sage: M.find_hexad([0, 1, 2, 4, 6])
            ([0, 1, 2, 4, 6, 8], ['line 1', 'picture 1'])
            sage: M = Minimog(type='modulo11')
            sage: M.find_hexad([1, 2, 3, 4, SR(infinity)]) # random (machine dependent?) order
            ([+Infinity, 2, 3, 4, 1, 10], ['square 8', 'picture 0'])

        AUTHOR:

        David Joyner (2006-05)

        REFERENCES: [Cu1984]_,  [Co1984]_
        """
        MINIMOG = self.minimog
        L = set(pts)
        LL = L.copy()
        pts_at_infty = {MINIMOG[0][2], MINIMOG[2][1], MINIMOG[0][0]}
        # recall & means intersection
        L2 = LL & pts_at_infty
        if len(L2) == 3:  # must be type 0 (line + pts at infty)
            H, WHAT = self.find_hexad0(LL - pts_at_infty)
            return H, WHAT
        if len(L2) == 2:  # type 0 or 3
            if (MINIMOG[0][2] in LL and MINIMOG[2][1] in LL):
                H, WHAT = self.find_hexad3(LL - {MINIMOG[0][2], MINIMOG[2][1]},
                                           MINIMOG[0][2], MINIMOG[2][1])
                if H:   # must be type 3
                    return list(H), WHAT
                # could be type 0
                H, WHAT = self.find_hexad0(LL - L2)
                if H:   # must be type 0
                    return list(H), WHAT
            if (MINIMOG[2][1] in LL and MINIMOG[0][0] in LL):
                H, WHAT = self.find_hexad3(LL - {MINIMOG[2][1], MINIMOG[0][0]},
                                           MINIMOG[2][1], MINIMOG[0][0])
                if H:   # must be type 3
                    return list(H), WHAT
                # could be type 0
                H, WHAT = self.find_hexad0(LL - L2)
                if H:   # must be type 0
                    return list(H), WHAT
            if (MINIMOG[0][2] in LL and MINIMOG[0][0] in LL):
                H, WHAT = self.find_hexad3(LL - {MINIMOG[0][2], MINIMOG[0][0]},
                                           MINIMOG[0][2], MINIMOG[0][0])
                if H:   # must be type 3
                    return list(H), WHAT
                # could be type 0
                H, WHAT = self.find_hexad0(LL - L2)
                if H:   # must be type 0
                    return list(H), WHAT
        if len(L2) == 1:
            H, WHAT = self.find_hexad2(LL - L2, list(L2)[0])
            if not H:   # not a cross in picture at infinity
                if list(L2)[0] == MINIMOG[2][1]:
                    L1 = LL - L2
                    H, WHAT = self.find_hexad3(L1, MINIMOG[0][0], MINIMOG[2][1])
                    if H:
                        return list(H), WHAT
                    L1 = LL - L2
                    H, WHAT = self.find_hexad3(L1, MINIMOG[0][2], MINIMOG[2][1])
                    if H:
                        return list(H), WHAT
                if list(L2)[0] == MINIMOG[0][0]:
                    L1 = (LL - L2)
                    H, WHAT = self.find_hexad3(L1, MINIMOG[0][0], MINIMOG[2][1])
                    if H:
                        return list(H), WHAT
                    L1 = (LL - L2)
                    H, WHAT = self.find_hexad3(L1, MINIMOG[0][0], MINIMOG[0][2])
                    if H:
                        return list(H), WHAT
                if list(L2)[0] == MINIMOG[0][2]:
                    L1 = (LL - L2)
                    H, WHAT = self.find_hexad3(L1, MINIMOG[0][0], MINIMOG[0][2])
                    if H:
                        return list(H), WHAT
                    L1 = (LL - L2)
                    H, WHAT = self.find_hexad3(L1, MINIMOG[2][1], MINIMOG[0][2])
                    if H:
                        return list(H), WHAT
            return list(H), WHAT
            # a cross in a pic at infty
        if not L2:  # L is either a union of 2 lines or a cross
            for i in LL:
                for j in pts_at_infty:
                    H, WHAT = self.find_hexad2(LL - {i}, j)
                    if (H and i in H):
                        return list(H), WHAT  # L is in a cross
            H, WHAT = self.find_hexad1(LL)  # L is a union of lines
            return H, WHAT

    def blackjack_move(self, L0):
        r"""
        Perform a blackjack move.

        INPUT:

        - ``L0`` -- list of cards of length 6, taken
          from `\{0, 1, ..., 11\}`

        .. RUBRIC:: MATHEMATICAL BLACKJACK

        Mathematical blackjack is played with 12 cards, labeled `0, ..., 11`
        (for example:  king,  ace, `2`, `3`, ..., `10`, jack, where the
        king is `0` and the  jack is `11`). Divide the 12 cards into two
        piles of `6` (to be fair, this should be done randomly). Each of
        the `6` cards of one of these piles are to be placed face up on
        the table. The remaining cards are in a stack which is shared
        and visible to both players. If the sum of the cards face up on
        the table is less than 21 then no legal move is possible so you
        must shuffle the cards and deal a new game. (Conway calls such
        a game `*={0|0}`, where `0={|}`; in this game the first player
        automatically wins.)

        * Players alternate moves.
        * A move consists of exchanging a card on the table with a
          lower card from the other pile.
        * The player whose move makes the sum of the cards on the table
          under 21 loses.

        The winning strategy (given below) for this game is due to
        Conway and Ryba. There is a Steiner system `S(5, 6, 12)` of hexads
        in the set `\{0, 1, ..., 11\}`. This Steiner system is associated
        to the MINIMOG of in the "shuffle numbering" rather than the
        "modulo `11` labeling".

        **Proposition** ([KR2001]_) For this Steiner system, the
        winning strategy is to choose a move which is a hexad from
        this system.

        EXAMPLES::

            sage: M = Minimog(type='modulo11')
            sage: M.blackjack_move([0, 2, 3, 6, 1, 10])
            '6 --> 5. The total went from 22 to 21.'
            sage: M = Minimog(type='shuffle')
            sage: M.blackjack_move([0, 2, 4, 6, 7, 11])
            '4 --> 3. The total went from 30 to 29.'

        Is this really a hexad? ::

            sage: M.find_hexad([11, 2, 3, 6, 7])
            ([0, 2, 3, 6, 7, 11], ['square 9', 'picture 1'])

        So, yes it is, but here is further confirmation::

            sage: M.blackjack_move([0, 2, 3, 6, 7, 11])
            This is a hexad.
            There is no winning move, so make a random legal move.
            [0, 2, 3, 6, 7, 11]

        Now, suppose player 2 replaced the 11 by a 9. Your next move::

            sage: M.blackjack_move([0, 2, 3, 6, 7, 9])
            '7 --> 1. The total went from 27 to 21.'

        You have now won. Sage will even tell you so::

            sage: M.blackjack_move([0, 2, 3, 6, 1, 9])
            'No move possible. Shuffle the deck and redeal.'

        AUTHOR:

        David Joyner (2006-05)

        REFERENCES: [CS1986]_, [KR2001]_
        """
        total = sum(L0)
        if total < 22:
            return "No move possible. Shuffle the deck and redeal."
        L = set(L0)
        for x in L:
            h, WHAT = self.find_hexad(L - {x})
            if list(L0) == list(h):
                print("      This is a hexad. \n      There is no winning move, so make a random legal move.")
                return L0
            y = list(set(h) - (L - {x}))[0]
            if y < x:
                return str(x) + ' --> ' + str(y) + ". The total went from " + str(total) + " to " + str(total - x + y) + "."
        print("This is a hexad. \n There is no winning move, so make a random legal move.")
        return L0



# sage_setup: distribution = sagemath-combinat
"""
Sum species
"""
#*****************************************************************************
#       Copyright (C) 2008 Mike Hansen <mhansen@gmail.com>,
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************
from .species import GenericCombinatorialSpecies
from .structure import SpeciesStructureWrapper
from sage.structure.unique_representation import UniqueRepresentation


class SumSpeciesStructure(SpeciesStructureWrapper):
    pass


class SumSpecies(GenericCombinatorialSpecies, UniqueRepresentation):
    def __init__(self, F, G, min=None, max=None, weight=None):
        """
        Return the sum of two species.

        EXAMPLES::

            sage: S = species.PermutationSpecies()
            sage: A = S+S
            sage: A.generating_series()[:5]
            [2, 2, 2, 2, 2]

            sage: P = species.PermutationSpecies()
            sage: F = P + P
            sage: F._check()                                                            # needs sage.libs.flint
            True
            sage: F == loads(dumps(F))
            True

        TESTS::

            sage: A = species.SingletonSpecies() + species.SingletonSpecies()
            sage: B = species.SingletonSpecies() + species.SingletonSpecies()
            sage: C = species.SingletonSpecies() + species.SingletonSpecies(min=2)
            sage: A is B
            True
            sage: (A is C) or (A == C)
            False
        """
        self._F = F
        self._G = G

        self._state_info = [F, G]

        GenericCombinatorialSpecies.__init__(self, min=None, max=None, weight=None)

    _default_structure_class = SumSpeciesStructure

    def left_summand(self):
        """
        Return the left summand of this species.

        EXAMPLES::

            sage: P = species.PermutationSpecies()
            sage: F = P + P*P
            sage: F.left_summand()
            Permutation species
        """
        return self._F

    def right_summand(self):
        """
        Return the right summand of this species.

        EXAMPLES::

            sage: P = species.PermutationSpecies()
            sage: F = P + P*P
            sage: F.right_summand()
            Product of (Permutation species) and (Permutation species)
        """
        return self._G

    def _name(self):
        """
        Note that we use a function to return the name of this species
        because we can't do it in the __init__ method due to it
        requiring that self.left_summand() and self.right_summand()
        already be unpickled.

        EXAMPLES::

            sage: P = species.PermutationSpecies()
            sage: F = P + P
            sage: F._name()
            'Sum of (Permutation species) and (Permutation species)'
        """
        return "Sum of (%s) and (%s)" % (self.left_summand(),
                                         self.right_summand())

    def _structures(self, structure_class, labels):
        """
        EXAMPLES::

            sage: P = species.PermutationSpecies()
            sage: F = P + P
            sage: F.structures([1,2]).list()
            [[1, 2], [2, 1], [1, 2], [2, 1]]
        """
        for res in self.left_summand().structures(labels):
            yield structure_class(self, res, tag='left')

        for res in self.right_summand().structures(labels):
            yield structure_class(self, res, tag='right')

    def _isotypes(self, structure_class, labels):
        """
        EXAMPLES::

            sage: P = species.PermutationSpecies()
            sage: F = P + P
            sage: F.isotypes([1,2]).list()                                              # needs sage.libs.flint
            [[2, 1], [1, 2], [2, 1], [1, 2]]
        """
        for res in self._F.isotypes(labels):
            yield structure_class(self, res, tag='left')

        for res in self._G.isotypes(labels):
            yield structure_class(self, res, tag='right')

    def _gs(self, series_ring, base_ring):
        """
        Return the cycle index series of this species.

        EXAMPLES::

            sage: P = species.PermutationSpecies()
            sage: F = P + P
            sage: F.generating_series()[:5]
            [2, 2, 2, 2, 2]
        """
        return (self.left_summand().generating_series(base_ring) +
                self.right_summand().generating_series(base_ring))

    def _itgs(self, series_ring, base_ring):
        """
        Return the isomorphism type generating series of this species.

        EXAMPLES::

            sage: P = species.PermutationSpecies()
            sage: F = P + P
            sage: F.isotype_generating_series()[:5]                                     # needs sage.libs.flint
            [2, 2, 4, 6, 10]
        """
        return (self.left_summand().isotype_generating_series(base_ring) +
                self.right_summand().isotype_generating_series(base_ring))

    def _cis(self, series_ring, base_ring):
        """
        Return the generating series of this species.

        EXAMPLES::

            sage: P = species.PermutationSpecies()
            sage: F = P + P
            sage: F.cycle_index_series()[:5]                                            # needs sage.modules
            [2*p[],
             2*p[1],
             2*p[1, 1] + 2*p[2],
             2*p[1, 1, 1] + 2*p[2, 1] + 2*p[3],
             2*p[1, 1, 1, 1] + 2*p[2, 1, 1] + 2*p[2, 2] + 2*p[3, 1] + 2*p[4]]
        """
        return (self.left_summand().cycle_index_series(base_ring) +
                self.right_summand().cycle_index_series(base_ring))

    def weight_ring(self):
        """
        Return the weight ring for this species. This is determined by
        asking Sage's coercion model what the result is when you add
        elements of the weight rings for each of the operands.

        EXAMPLES::

            sage: S = species.SetSpecies()
            sage: C = S+S
            sage: C.weight_ring()
            Rational Field

        ::

            sage: S = species.SetSpecies(weight=QQ['t'].gen())
            sage: C = S + S
            sage: C.weight_ring()
            Univariate Polynomial Ring in t over Rational Field
        """
        return self._common_parent([self.left_summand().weight_ring(),
                                    self.right_summand().weight_ring()])

    def _equation(self, var_mapping):
        """
        Return the right hand side of an algebraic equation satisfied by
        this species. This is a utility function called by the
        algebraic_equation_system method.

        EXAMPLES::

            sage: X = species.SingletonSpecies()
            sage: S = X + X
            sage: S.algebraic_equation_system()                                         # needs sage.graphs
            [node1 + (-2*z)]
        """
        return sum(var_mapping[operand] for operand in self._state_info)


#Backward compatibility
SumSpecies_class = SumSpecies



# sage_setup: distribution = sagemath-combinat
# sage.doctest: needs sage.combinat sage.modules
"""
Finite dimensional free algebra quotients

REMARK:

This implementation only works for finite dimensional quotients, since
a list of basis monomials and the multiplication matrices need to be
explicitly provided.

The homogeneous part of a quotient of a free algebra over a field by a
finitely generated homogeneous twosided ideal is available in a
different implementation. See
:mod:`~sage.algebras.letterplace.free_algebra_letterplace` and
:mod:`~sage.rings.quotient_ring`.

TESTS:

::

    sage: n = 2
    sage: A = FreeAlgebra(QQ,n,'x')
    sage: F = A.monoid()
    sage: i, j = F.gens()
    sage: mons = [F(1), i, j, i*j]
    sage: r = len(mons)
    sage: M = MatrixSpace(QQ,r)
    sage: mats = [M([0,1,0,0, -1,0,0,0, 0,0,0,-1, 0,0,1,0]),
    ....:         M([0,0,1,0, 0,0,0,1, -1,0,0,0, 0,-1,0,0]) ]
    sage: H2.<i,j> = A.quotient(mons, mats)
    sage: H2 == loads(dumps(H2))
    True
    sage: i == loads(dumps(i))
    True

Test comparison by equality::

    sage: HQ = sage.algebras.free_algebra_quotient.hamilton_quatalg(QQ)[0]
    sage: HZ = sage.algebras.free_algebra_quotient.hamilton_quatalg(ZZ)[0]
    sage: HQ == HQ
    True
    sage: HQ == HZ
    False
    sage: HZ == QQ
    False
"""

# ****************************************************************************
#  Copyright (C) 2005 David Kohel <kohel@maths.usyd.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty
#    of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License for more details; the full text
#  is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.algebras.free_algebra import FreeAlgebra_generic
from sage.algebras.free_algebra_quotient_element import FreeAlgebraQuotientElement
from sage.categories.algebras import Algebras
from sage.misc.lazy_import import lazy_import
from sage.modules.free_module import FreeModule
from sage.structure.unique_representation import UniqueRepresentation
from sage.structure.parent import Parent

lazy_import('sage.algebras.letterplace.free_algebra_letterplace', 'FreeAlgebra_letterplace')


class FreeAlgebraQuotient(UniqueRepresentation, Parent):
    @staticmethod
    def __classcall__(cls, A, mons, mats, names):
        """
        Used to support unique representation.

        EXAMPLES::

            sage: H = sage.algebras.free_algebra_quotient.hamilton_quatalg(QQ)[0]  # indirect doctest
            sage: H1 = sage.algebras.free_algebra_quotient.hamilton_quatalg(QQ)[0]
            sage: H is H1
            True
        """
        new_mats = []
        for M in mats:
            M = M.parent()(M)
            M.set_immutable()
            new_mats.append(M)
        return super().__classcall__(cls, A, tuple(mons),
                                     tuple(new_mats), tuple(names))

    Element = FreeAlgebraQuotientElement

    def __init__(self, A, mons, mats, names):
        """
        Return a quotient algebra defined via the action of a free algebra
        A on a (finitely generated) free module.

        The input for the quotient algebra is a list of monomials (in
        the underlying monoid for A) which form a free basis for the
        module of A, and a list of matrices, which give the action of
        the free generators of A on this monomial basis.

        EXAMPLES:

        Quaternion algebra defined in terms of three generators::

            sage: n = 3
            sage: A = FreeAlgebra(QQ,n,'i')
            sage: F = A.monoid()
            sage: i, j, k = F.gens()
            sage: mons = [F(1), i, j, k]
            sage: M = MatrixSpace(QQ,4)
            sage: mats = [M([0,1,0,0, -1,0,0,0, 0,0,0,-1, 0,0,1,0]),
            ....:         M([0,0,1,0, 0,0,0,1, -1,0,0,0, 0,-1,0,0]),
            ....:         M([0,0,0,1, 0,0,-1,0, 0,1,0,0, -1,0,0,0]) ]
            sage: H3.<i,j,k> = FreeAlgebraQuotient(A,mons,mats)
            sage: x = 1 + i + j + k
            sage: x
            1 + i + j + k
            sage: x**128
            -170141183460469231731687303715884105728
             + 170141183460469231731687303715884105728*i
             + 170141183460469231731687303715884105728*j
             + 170141183460469231731687303715884105728*k

        Same algebra defined in terms of two generators, with some penalty
        on already slow arithmetic.

        ::

            sage: n = 2
            sage: A = FreeAlgebra(QQ,n,'x')
            sage: F = A.monoid()
            sage: i, j = F.gens()
            sage: mons = [ F(1), i, j, i*j ]
            sage: r = len(mons)
            sage: M = MatrixSpace(QQ,r)
            sage: mats = [M([0,1,0,0, -1,0,0,0, 0,0,0,-1, 0,0,1,0]),
            ....:         M([0,0,1,0, 0,0,0,1, -1,0,0,0, 0,-1,0,0]) ]
            sage: H2.<i,j> = A.quotient(mons,mats)
            sage: k = i*j
            sage: x = 1 + i + j + k
            sage: x
            1 + i + j + i*j
            sage: x**128
            -170141183460469231731687303715884105728
             + 170141183460469231731687303715884105728*i
             + 170141183460469231731687303715884105728*j
             + 170141183460469231731687303715884105728*i*j

        TESTS::

            sage: TestSuite(H2).run()
        """
        if not isinstance(A, (FreeAlgebra_generic, FreeAlgebra_letterplace)):
            raise TypeError("argument A must be a free algebra")
        R = A.base_ring()
        n = A.ngens()
        assert n == len(mats)
        self.__free_algebra = A
        self.__ngens = n
        self.__dim = len(mons)
        self.__module = FreeModule(R, self.__dim)
        self.__matrix_action = mats
        self.__monomial_basis = mons  # elements of free monoid
        Parent.__init__(self, base=R, names=names,
                        normalize=True, category=Algebras(R))

    def _element_constructor_(self, x):
        """
        EXAMPLES::

            sage: H, (i,j,k) = sage.algebras.free_algebra_quotient.hamilton_quatalg(QQ)
            sage: H(i) is i
            True
            sage: a = H._element_constructor_(1); a
            1
            sage: a in H
            True
            sage: a = H._element_constructor_([1,2,3,4]); a
            1 + 2*i + 3*j + 4*k
        """
        return self.element_class(self, x)

    def _coerce_map_from_(self, S):
        """
        EXAMPLES::

            sage: H, (i,j,k) = sage.algebras.free_algebra_quotient.hamilton_quatalg(QQ)
            sage: H._coerce_map_from_(H)
            True
            sage: H._coerce_map_from_(QQ)
            True
            sage: H._coerce_map_from_(GF(7))
            False
        """
        return S == self or self.__free_algebra.has_coerce_map_from(S)

    def _repr_(self) -> str:
        """
        EXAMPLES::

            sage: H, (i,j,k) = sage.algebras.free_algebra_quotient.hamilton_quatalg(QQ)
            sage: H._repr_()
            "Free algebra quotient on 3 generators ('i', 'j', 'k') and dimension 4 over Rational Field"
        """
        R = self.base_ring()
        n = self.__ngens
        r = self.__module.dimension()
        x = self.variable_names()
        return f"Free algebra quotient on {n} generators {x} and dimension {r} over {R}"

    def gen(self, i):
        """
        Return the ``i``-th generator of the algebra.

        EXAMPLES::

            sage: H, (i,j,k) = sage.algebras.free_algebra_quotient.hamilton_quatalg(QQ)
            sage: H.gen(0)
            i
            sage: H.gen(2)
            k

        An :exc:`IndexError` is raised if an invalid generator is requested::

            sage: H.gen(3)
            Traceback (most recent call last):
            ...
            IndexError: argument i (= 3) must be between 0 and 2

        Negative indexing into the generators is not supported::

            sage: H.gen(-1)
            Traceback (most recent call last):
            ...
            IndexError: argument i (= -1) must be between 0 and 2
        """
        n = self.__ngens
        if i < 0 or not i < n:
            raise IndexError(f"argument i (= {i}) must be between 0 and {n - 1}")
        one = self.base_ring().one()
        F = self.__free_algebra.monoid()
        return self.element_class(self, {F.gen(i): one})

    def gens(self) -> tuple:
        """
        Return the tuple of generators of ``self``.

        EXAMPLES::

            sage: H, (i,j,k) = sage.algebras.free_algebra_quotient.hamilton_quatalg(QQ)
            sage: H.gens()
            (i, j, k)
        """
        one = self.base_ring().one()
        F = self.__free_algebra.monoid()
        return tuple(self.element_class(self, {F.gen(i): one})
                     for i in range(self.__ngens))

    def ngens(self):
        """
        Return the number of generators of the algebra.

        EXAMPLES::

            sage: sage.algebras.free_algebra_quotient.hamilton_quatalg(QQ)[0].ngens()
            3
        """
        return self.__ngens

    def dimension(self):
        """
        Return the rank of the algebra (as a free module).

        EXAMPLES::

            sage: sage.algebras.free_algebra_quotient.hamilton_quatalg(QQ)[0].dimension()
            4
        """
        return self.__dim

    def matrix_action(self):
        """
        Return the matrix action used to define the algebra.

        EXAMPLES::

            sage: sage.algebras.free_algebra_quotient.hamilton_quatalg(QQ)[0].matrix_action()
            (
            [ 0  1  0  0]  [ 0  0  1  0]  [ 0  0  0  1]
            [-1  0  0  0]  [ 0  0  0  1]  [ 0  0 -1  0]
            [ 0  0  0 -1]  [-1  0  0  0]  [ 0  1  0  0]
            [ 0  0  1  0], [ 0 -1  0  0], [-1  0  0  0]
            )
        """
        return self.__matrix_action

    def monomial_basis(self):
        """
        The free monoid of generators of the algebra as elements of a free
        monoid.

        EXAMPLES::

            sage: sage.algebras.free_algebra_quotient.hamilton_quatalg(QQ)[0].monomial_basis()
            (1, i0, i1, i2)
        """
        return self.__monomial_basis

    def rank(self):
        """
        Return the rank of the algebra (as a free module).

        EXAMPLES::

            sage: sage.algebras.free_algebra_quotient.hamilton_quatalg(QQ)[0].rank()
            4
        """
        return self.__dim

    def module(self):
        """
        Return the free module of the algebra.

        EXAMPLES::

            sage: H = sage.algebras.free_algebra_quotient.hamilton_quatalg(QQ)[0]; H
            Free algebra quotient on 3 generators ('i', 'j', 'k') and dimension 4 over Rational Field
            sage: H.module()
            Vector space of dimension 4 over Rational Field
        """
        return self.__module

    def monoid(self):
        """
        Return the free monoid of generators of the algebra.

        EXAMPLES::

            sage: sage.algebras.free_algebra_quotient.hamilton_quatalg(QQ)[0].monoid()
            Free monoid on 3 generators (i0, i1, i2)
        """
        return self.__free_algebra.monoid()

    def free_algebra(self):
        """
        Return the free algebra generating the algebra.

        EXAMPLES::

            sage: sage.algebras.free_algebra_quotient.hamilton_quatalg(QQ)[0].free_algebra()
            Free Algebra on 3 generators (i0, i1, i2) over Rational Field
        """
        return self.__free_algebra


def hamilton_quatalg(R):
    """
    Hamilton quaternion algebra over the commutative ring ``R``,
    constructed as a free algebra quotient.

    INPUT:

    - ``R`` -- a commutative ring

    OUTPUT:

    - ``Q`` -- quaternion algebra
    - ``gens`` -- generators for ``Q``

    EXAMPLES::

        sage: H, (i,j,k) = sage.algebras.free_algebra_quotient.hamilton_quatalg(ZZ)
        sage: H
        Free algebra quotient on 3 generators ('i', 'j', 'k') and dimension 4
         over Integer Ring
        sage: i^2
        -1
        sage: i in H
        True

    Note that there is another vastly more efficient model for
    quaternion algebras in Sage; the one here is mainly for testing
    purposes::

        sage: R.<i,j,k> = QuaternionAlgebra(QQ,-1,-1)  # much fast than the above       # needs sage.libs.singular
    """
    from sage.algebras.free_algebra import FreeAlgebra
    from sage.matrix.matrix_space import MatrixSpace
    A = FreeAlgebra(R, 3, 'i')
    F = A.monoid()
    i, j, k = F.gens()
    mons = [F.one(), i, j, k]
    M = MatrixSpace(R, 4)
    mats = [M([0, 1, 0, 0, -1, 0, 0, 0, 0, 0, 0, -1, 0, 0, 1, 0]),
            M([0, 0, 1, 0, 0, 0, 0, 1, -1, 0, 0, 0, 0, -1, 0, 0]),
            M([0, 0, 0, 1, 0, 0, -1, 0, 0, 1, 0, 0, -1, 0, 0, 0])]
    H3 = FreeAlgebraQuotient(A, mons, mats, names=('i', 'j', 'k'))
    return H3, H3.gens()



# sage_setup: distribution = sagemath-combinat
# sage.doctest: needs sage.combinat sage.modules
r"""
LLT symmetric functions

REFERENCES:

.. [LLT1997] Alain Lascoux, Bernard Leclerc, Jean-Yves Thibon,
   Ribbon tableaux, Hall-Littlewood functions, quantum affine algebras, and unipotent varieties,
   J. Math. Phys. 38 (1997), no. 2, 1041-1068,
   :arxiv:`q-alg/9512031v1` [math.q.alg]

.. [LT2000] Bernard Leclerc and Jean-Yves Thibon,
   Littlewood-Richardson coefficients and Kazhdan-Lusztig polynomials,
   in: Combinatorial methods in representation theory (Kyoto)
   Adv. Stud. Pure Math., vol. 28, Kinokuniya, Tokyo, 2000, pp 155-220
   :arxiv:`math/9809122v3` [math.q-alg]
"""
# ****************************************************************************
#       Copyright (C) 2007 Mike Hansen <mhansen@gmail.com>
#                     2012 Mike Zabrocki <mike.zabrocki@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************
import sage.combinat.skew_partition
from sage.categories.homset import Hom
from sage.categories.morphism import SetMorphism
from sage.combinat import ribbon_tableau
from sage.combinat.partition import Partition, Partitions, _Partitions
from sage.rings.integer_ring import ZZ
from sage.rings.rational_field import QQ
from sage.structure.unique_representation import UniqueRepresentation

from . import sfa

# cache for H spin basis
hsp_to_m_cache = {}
m_to_hsp_cache = {}

# cache for H cospin basis
hcosp_to_m_cache = {}
m_to_hcosp_cache = {}

QQt = QQ['t'].fraction_field()
# This is to become the "abstract algebra" for llt polynomials


class LLT_class(UniqueRepresentation):
    r"""
    A class for working with LLT symmetric functions.

    EXAMPLES::

        sage: Sym = SymmetricFunctions(FractionField(QQ['t']))
        sage: L3 = Sym.llt(3); L3
        level 3 LLT polynomials over Fraction Field of Univariate Polynomial Ring in t over Rational Field
        sage: L3.cospin([3,2,1])
        (t+1)*m[1, 1] + m[2]
        sage: HC3 = L3.hcospin(); HC3
        Symmetric Functions over Fraction Field of Univariate Polynomial Ring in t over Rational Field in the level 3 LLT cospin basis
        sage: m = Sym.monomial()
        sage: m( HC3[1,1] )
        (t+1)*m[1, 1] + m[2]

    We require that the parameter `t` must be in the base ring::

        sage: # needs sage.symbolic
        sage: Symxt = SymmetricFunctions(QQ['x','t'].fraction_field())
        sage: (x,t) = Symxt.base_ring().gens()
        sage: LLT3x = Symxt.llt(3,t=x)
        sage: LLT3 = Symxt.llt(3)
        sage: HS3x = LLT3x.hspin()
        sage: HS3t = LLT3.hspin()
        sage: s = Symxt.schur()
        sage: s(HS3x[2,1])
        s[2, 1] + x*s[3]
        sage: s(HS3t[2,1])
        s[2, 1] + t*s[3]
        sage: HS3x(HS3t[2,1])
        HSp3[2, 1] - (x-t)*HSp3[3]
        sage: s(HS3x(HS3t[2,1]))
        s[2, 1] + t*s[3]
        sage: LLT3t2 = Symxt.llt(3,t=2)
        sage: HC3t2 = LLT3t2.hcospin()
        sage: HS3x(HC3t2[3,1])
        2*HSp3[3, 1] - (2*x-1)*HSp3[4]
    """
    @staticmethod
    def __classcall__(cls, Sym, k, t='t'):
        """
        Normalize the arguments.

        TESTS::

            sage: R.<q, t> = QQ[]
            sage: B1 = SymmetricFunctions(R).llt(3).hspin()
            sage: B2 = SymmetricFunctions(R).llt(3, t).hspin()
            sage: B3 = SymmetricFunctions(R).llt(3, q).hspin()
            sage: B1 is B2
            True
            sage: B1 == B3
            False
        """
        return super().__classcall__(cls, Sym, k, Sym.base_ring()(t))

    def __init__(self, Sym, k, t):
        r"""
        Class of LLT symmetric function bases.

        INPUT:

        - ``self`` -- a family of LLT symmetric function bases
        - ``k`` -- positive integer (the level)
        - ``t`` -- a parameter (default: `t`)

        EXAMPLES::

            sage: L3 = SymmetricFunctions(FractionField(QQ['t'])).llt(3)
            sage: L3 == loads(dumps(L3))
            True
            sage: TestSuite(L3).run(skip=["_test_associativity","_test_distributivity","_test_prod"])

        TESTS::

            sage: L3 != SymmetricFunctions(FractionField(QQ['t'])).llt(2)
            True
            sage: L3p = SymmetricFunctions(FractionField(QQ['t'])).llt(3,t=1)
            sage: L3 != L3p
            True
            sage: L3p != SymmetricFunctions(QQ).llt(3,t=1)
            True

            sage: Sym = SymmetricFunctions(QQ['t'])
            sage: ks3 = Sym.kschur(3)
            sage: llt3 = Sym.llt(3)
            sage: f = llt3.cospin([[1],[2,1],[1,1]]).omega()
            sage: ks3(f)
            ks3[2, 2, 1, 1] + ks3[2, 2, 2] + t*ks3[3, 1, 1, 1] + t^2*ks3[3, 2, 1]
        """
        self._k = k
        self._sym = Sym
        self._name = "level %s LLT polynomials" % self._k
        self.t = t
        self._name_suffix = ""
        if str(t) != 't':
            self._name_suffix += " with t=%s" % self.t
        self._name += self._name_suffix + " over %s" % self._sym.base_ring()
        self._m = Sym.monomial()

    def __repr__(self):
        r"""
        Representation of the LLT symmetric functions.

        INPUT:

        - ``self`` -- a family of LLT symmetric function bases

        OUTPUT: a string representing the LLT symmetric functions

        EXAMPLES::

            sage: SymmetricFunctions(FractionField(QQ['t'])).llt(3)
            level 3 LLT polynomials over Fraction Field of Univariate Polynomial Ring in t over Rational Field
            sage: SymmetricFunctions(QQ).llt(3,t=2)
            level 3 LLT polynomials with t=2 over Rational Field
        """
        return self._name

    def symmetric_function_ring( self ):
        r"""
        The symmetric function algebra associated to the family of LLT
        symmetric function bases

        INPUT:

        - ``self`` -- a family of LLT symmetric functions bases

        OUTPUT: the symmetric function ring associated to ``self``

        EXAMPLES::

            sage: L3 = SymmetricFunctions(FractionField(QQ['t'])).llt(3)
            sage: L3.symmetric_function_ring()
            Symmetric Functions over Fraction Field of Univariate Polynomial Ring in t over Rational Field
        """
        return self._sym

    def base_ring(self):
        r"""
        Return the base ring of ``self``.

        INPUT:

        - ``self`` -- a family of LLT symmetric functions bases

        OUTPUT: the base ring of the symmetric function ring associated to ``self``

        EXAMPLES::

            sage: SymmetricFunctions(FractionField(QQ['t'])).llt(3).base_ring()
            Fraction Field of Univariate Polynomial Ring in t over Rational Field
        """
        return self._sym.base_ring()

    def level(self):
        r"""
        Return the level of ``self``.

        INPUT:

        - ``self`` -- a family of LLT symmetric functions bases

        OUTPUT: the level is the parameter of `k` in the basis

        EXAMPLES::

            sage: SymmetricFunctions(FractionField(QQ['t'])).llt(3).level()
            3
        """
        return self._k

    def _llt_generic(self, skp, stat):
        r"""
        Take in partition, list of partitions, or a list of skew
        partitions as well as a function which takes in two partitions and
        a level and return a coefficient.

        INPUT:

        - ``self`` -- a family of LLT symmetric functions bases
        - ``skp`` -- a partition or a list of partitions or a list of skew partitions
        - ``stat`` -- a function which accepts two partitions and a value
          for the level and returns a coefficient which is a polynomial
          in a parameter `t`.  The first partition is the index of the
          LLT function, the second partition is the index of the monomial
          basis element.

        OUTPUT:

        - returns the monomial expansion of the LLT symmetric function
          indexed by ``skp``

        EXAMPLES::

            sage: L3 = SymmetricFunctions(FractionField(QQ['t'])).llt(3)
            sage: f = lambda skp,mu,level: QQ(1)
            sage: L3._llt_generic([3,2,1],f)
            m[1, 1] + m[2]
            sage: L3._llt_generic([[2,1],[1],[2]],f)
            m[1, 1, 1, 1, 1, 1] + m[2, 1, 1, 1, 1] + m[2, 2, 1, 1] + m[2, 2, 2] + m[3, 1, 1, 1] + m[3, 2, 1] + m[3, 3] + m[4, 1, 1] + m[4, 2] + m[5, 1] + m[6]
            sage: L3._llt_generic([[[2,2],[1]],[[2,1],[]]],f)
            m[1, 1, 1, 1] + m[2, 1, 1] + m[2, 2] + m[3, 1] + m[4]
        """
        if skp in _Partitions:
            m = (sum(skp) / self.level()).floor()
            if m == 0:
                raise ValueError("level (%s=) must divide %s " % (sum(skp),
                                                                  self.level()))
            mu = Partitions( ZZ(sum(skp) / self.level()) )

        elif isinstance(skp, list) and skp[0] in sage.combinat.skew_partition.SkewPartitions():
            #skp is a list of skew partitions
            skp2 = [Partition(core=[], quotient=[skp[i][0] for i in range(len(skp))])]
            skp2 += [Partition(core=[], quotient=[skp[i][1] for i in range(len(skp))])]
            mu = Partitions(ZZ((skp2[0].size()-skp2[1].size()) / self.level()))
            skp = skp2
        elif isinstance(skp, list) and skp[0] in _Partitions:
            #skp is a list of partitions
            skp = Partition(core=[], quotient=skp)
            mu = Partitions( ZZ(sum(skp) / self.level()) )
        else:
            raise ValueError("LLT polynomials not defined for %s" % skp)

        BR = self.base_ring()
        return sum([ BR(stat(skp,nu,self.level()).subs(t=self.t))*self._m(nu) for nu in mu])

    def spin_square(self, skp):
        r"""
        Calculate a single instance of a spin squared LLT symmetric function
        associated with a partition, list of partitions, or a list of skew partitions.

        This family of symmetric functions is defined in [LT2000]_ equation (43).

        INPUT:

        - ``self`` -- a family of LLT symmetric functions bases
        - ``skp`` -- a partition of a list of partitions or a list of skew
          partitions

        OUTPUT:

        the monomial expansion of the LLT symmetric function spin-square
        functions indexed by ``skp``

        EXAMPLES::

            sage: L3 = SymmetricFunctions(FractionField(QQ['t'])).llt(3)
            sage: L3.spin_square([2,1])
            t*m[1]
            sage: L3.spin_square([3,2,1])
            (t^3+t)*m[1, 1] + t^3*m[2]
            sage: L3.spin_square([[1],[1],[1]])
            (t^6+2*t^4+2*t^2+1)*m[1, 1, 1] + (t^6+t^4+t^2)*m[2, 1] + t^6*m[3]
            sage: L3.spin_square([[[2,2],[1]],[[2,1],[]]])
            (2*t^4+3*t^2+1)*m[1, 1, 1, 1] + (t^4+t^2)*m[2, 1, 1] + t^4*m[2, 2]
        """
        return self._llt_generic(skp, ribbon_tableau.spin_polynomial_square)

    def cospin(self, skp):
        r"""
        Calculate a single instance of the cospin symmetric functions.

        These are the functions defined in [LLT1997]_ equation (26).

        INPUT:

        - ``self`` -- a family of LLT symmetric functions bases
        - ``skp`` -- a partition or a list of partitions or a list of skew
          partitions

        OUTPUT:

        the monomial expansion of the LLT symmetric function cospin
        functions indexed by ``skp``

        EXAMPLES::

            sage: Sym = SymmetricFunctions(FractionField(QQ['t']))
            sage: L3 = Sym.llt(3)
            sage: L3.cospin([2,1])
            m[1]
            sage: L3.cospin([3,2,1])
            (t+1)*m[1, 1] + m[2]
            sage: s = Sym.schur()
            sage: s(L3.cospin([[2],[1],[2]]))
            t^4*s[2, 2, 1] + t^3*s[3, 1, 1] + (t^3+t^2)*s[3, 2] + (t^2+t)*s[4, 1] + s[5]
        """
        return self._llt_generic(skp, ribbon_tableau.cospin_polynomial)

#### Is it safe to delete this function?
##     def llt_inv(self, skp):
##         """
##         """
##         l = sage.combinat.partitions( sum( [ p.size() for p in skp ] ) ).list()
##         res = m(0)
##         for p in l:
##             inv_p = [ ktuple.inversions() for ktuple in kTupleTableaux(skp, p) ]
##             res += sum([t**x for x in inv_p])*m(p)
##         return res

    def hcospin(self):
        r"""
        Return the HCospin basis.
        This basis is defined [LLT1997]_ equation (27).

        INPUT:

        - ``self`` -- a family of LLT symmetric functions bases

        OUTPUT: the h-cospin basis of the LLT symmetric functions

        EXAMPLES::

            sage: Sym = SymmetricFunctions(FractionField(QQ['t']))
            sage: HCosp3 = Sym.llt(3).hcospin(); HCosp3
            Symmetric Functions
             over Fraction Field of Univariate Polynomial Ring in t over Rational Field
             in the level 3 LLT cospin basis
            sage: HCosp3([1])^2
            1/t*HCosp3[1, 1] + ((t-1)/t)*HCosp3[2]

            sage: s = Sym.schur()
            sage: HCosp3(s([2]))
            HCosp3[2]
            sage: HCosp3(s([1,1]))
            1/t*HCosp3[1, 1] - 1/t*HCosp3[2]
            sage: s(HCosp3([2,1]))
            t*s[2, 1] + s[3]
        """
        return LLT_cospin(self)

    def hspin(self):
        r"""
        Return the HSpin basis.

        This basis is defined [LLT1997]_ equation (28).

        INPUT:

        - ``self`` -- a family of LLT symmetric functions bases

        OUTPUT: the h-spin basis of the LLT symmetric functions

        EXAMPLES::

            sage: Sym = SymmetricFunctions(FractionField(QQ['t']))

            sage: # needs sage.symbolic
            sage: HSp3 = Sym.llt(3).hspin(); HSp3
            Symmetric Functions
             over Fraction Field of Univariate Polynomial Ring in t over Rational Field
             in the level 3 LLT spin basis
            sage: HSp3([1])^2
            HSp3[1, 1] + (-t+1)*HSp3[2]
            sage: s = Sym.schur()
            sage: HSp3(s([2]))
            HSp3[2]
            sage: HSp3(s([1,1]))
            HSp3[1, 1] - t*HSp3[2]
            sage: s(HSp3([2,1]))
            s[2, 1] + t*s[3]
        """
        return LLT_spin(self)


class LLT_generic(sfa.SymmetricFunctionAlgebra_generic):

    def __init__(self, llt, prefix):
        r"""
        A class of methods which are common to both the hspin and hcospin
        of the LLT symmetric functions.

        INPUT:

        - ``self`` -- an instance of the LLT hspin or hcospin basis
        - ``llt`` -- a family of LLT symmetric functions

        EXAMPLES::

            sage: SymmetricFunctions(FractionField(QQ['t'])).llt(3).hspin()
            Symmetric Functions
             over Fraction Field of Univariate Polynomial Ring in t over Rational Field
             in the level 3 LLT spin basis
            sage: SymmetricFunctions(QQ).llt(3,t=2).hspin()
            Symmetric Functions over Rational Field in the level 3 LLT spin with t=2 basis
            sage: QQz = FractionField(QQ['z']); z = QQz.gen()
            sage: SymmetricFunctions(QQz).llt(3,t=z).hspin()
            Symmetric Functions
             over Fraction Field of Univariate Polynomial Ring in z over Rational Field
             in the level 3 LLT spin with t=z basis
        """
        s = self.__class__.__name__[4:]
        sfa.SymmetricFunctionAlgebra_generic.__init__(
            self, llt._sym,
            basis_name="level %s LLT " % llt.level() + s + llt._name_suffix,
            prefix=prefix)

        self.t = llt.t
        self._sym = llt._sym
        self._llt = llt
        self._k = llt._k

        sfa.SymmetricFunctionAlgebra_generic.__init__(self, self._sym, self._basis_name)

        # temporary until Hom(GradedHopfAlgebrasWithBasis work better)
        category = sage.categories.all.ModulesWithBasis(self._sym.base_ring())
        self._m = llt._sym.m()
        self   .register_coercion(SetMorphism(Hom(self._m, self, category), self._m_to_self))
        self._m.register_coercion(SetMorphism(Hom(self, self._m, category), self._self_to_m))

    def construction(self):
        """
        Return a pair ``(F, R)``, where ``F`` is a
        :class:`SymmetricFunctionsFunctor` and `R` is a ring, such
        that ``F(R)`` returns ``self``.

        EXAMPLES::

            sage: Sym = SymmetricFunctions(FractionField(QQ['t']))
            sage: HSp3 = Sym.llt(3).hspin()
            sage: HSp3.construction()
            (SymmetricFunctionsFunctor[level 3 LLT spin],
             Fraction Field of Univariate Polynomial Ring in t over Rational Field)
        """
        from sage.combinat.sf.sfa import SymmetricFunctionsFamilyFunctor
        return (SymmetricFunctionsFamilyFunctor(self, LLT_class,
                                                self.basis_name(),
                                                self._k, self.t),
                self.base_ring())

    def _m_to_self(self, x):
        r"""
        Isomorphism from the monomial basis into ``self``.

        INPUT:

        - ``self`` -- an instance of the LLT hspin or hcospin basis
        - ``x`` -- an element of the monomial basis

        OUTPUT: ``x`` expanded in the basis ``self``

        EXAMPLES::

            sage: # needs sage.symbolic
            sage: Sym = SymmetricFunctions(FractionField(QQ['t']))
            sage: HSp3 = Sym.llt(3).hspin()
            sage: m = Sym.monomial()
            sage: HSp3._m_to_self(m[2,1])
            -2*HSp3[1, 1, 1] + (2*t^2+2*t+1)*HSp3[2, 1] + (-2*t^2-t)*HSp3[3]

        This is for internal use only. Please use instead::

            sage: HSp3(m[2,1])                                                          # needs sage.symbolic
            -2*HSp3[1, 1, 1] + (2*t^2+2*t+1)*HSp3[2, 1] + (-2*t^2-t)*HSp3[3]
        """
        return self._from_cache(x, self._m_cache, self._m_to_self_cache,
                                t=self.t)

    def _self_to_m(self, x):
        r"""
        Isomorphism from ``self`` to the monomial basis.

        INPUT:

        - ``self`` -- an instance of the LLT hspin or hcospin basis
        - ``x`` -- an element of ``self``

        OUTPUT: ``x`` expanded in the monomial basis

        EXAMPLES::

            sage: # needs sage.symbolic
            sage: Sym = SymmetricFunctions(FractionField(QQ['t']))
            sage: HSp3 = Sym.llt(3).hspin()
            sage: m = Sym.monomial()
            sage: HSp3._self_to_m(HSp3[2,1])
            (t+2)*m[1, 1, 1] + (t+1)*m[2, 1] + t*m[3]

        This is for internal use only. Please use instead::

            sage: m(HSp3[2,1])                                                          # needs sage.symbolic
            (t+2)*m[1, 1, 1] + (t+1)*m[2, 1] + t*m[3]
        """
        return self._m._from_cache(x, self._m_cache, self._self_to_m_cache,
                                   t=self.t)

    def level(self):
        r"""
        Return the level of ``self``.

        INPUT:

        - ``self`` -- an instance of the LLT hspin or hcospin basis

        OUTPUT: the level associated to the basis ``self``

        EXAMPLES::

            sage: HSp3 = SymmetricFunctions(FractionField(QQ['t'])).llt(3).hspin()
            sage: HSp3.level()
            3
        """
        return self._k

    def llt_family( self ):
        r"""
        The family of the llt bases of the symmetric functions.

        INPUT:

        - ``self`` -- an instance of the LLT hspin or hcospin basis

        OUTPUT: an instance of the family of LLT bases associated to ``self``

        EXAMPLES::

            sage: HSp3 = SymmetricFunctions(FractionField(QQ['t'])).llt(3).hspin()
            sage: HSp3.llt_family()
            level 3 LLT polynomials
             over Fraction Field of Univariate Polynomial Ring in t over Rational Field
        """
        return self._llt

    def product(self, left, right):
        r"""
        Convert to the monomial basis, do the multiplication there, and
        convert back to the basis ``self``.

        INPUT:

        - ``self`` -- an instance of the LLT hspin or hcospin basis
        - ``left``, ``right`` -- elements of the symmetric functions

        OUTPUT: the product of ``left`` and ``right`` expanded in the basis ``self``

        EXAMPLES::

            sage: # needs sage.symbolic
            sage: HSp3 = SymmetricFunctions(FractionField(QQ['t'])).llt(3).hspin()
            sage: HSp3.product(HSp3([1]), HSp3([2]))
            HSp3[2, 1] + (-t+1)*HSp3[3]
            sage: HCosp3 = SymmetricFunctions(FractionField(QQ['t'])).llt(3).hcospin()
            sage: HCosp3.product(HCosp3([1]), HSp3([2]))
            1/t*HCosp3[2, 1] + ((t-1)/t)*HCosp3[3]
        """
        return self(self._m(left) * self._m(right))

    def _m_cache(self, n):
        r"""
        Compute the change of basis from the monomial symmetric functions
        to ``self``.

        INPUT:

        - ``self`` -- an instance of the LLT hspin or hcospin basis
        - ``n`` -- positive integer representing the degree

        EXAMPLES::

            sage: # needs sage.symbolic
            sage: HSp3 = SymmetricFunctions(FractionField(QQ['t'])).llt(3).hspin()
            sage: HSp3._m_cache(2)
            sage: l = lambda c: [ (i[0],[j for j in sorted(i[1].items())]) for i in sorted(c.items())]
            sage: l( HSp3._self_to_m_cache[2] )
            [([1, 1], [([1, 1], t + 1), ([2], t)]), ([2], [([1, 1], 1), ([2], 1)])]
            sage: l( HSp3._m_to_self_cache[2] )
            [([1, 1], [([1, 1], 1), ([2], -t)]), ([2], [([1, 1], -1), ([2], t + 1)])]
            sage: HCosp3 = SymmetricFunctions(FractionField(QQ['t'])).llt(3).hcospin()
            sage: HCosp3._m_cache(2)
            sage: l = lambda c: [ (i[0],[j for j in sorted(i[1].items())]) for i in sorted(c.items())]
            sage: l( HCosp3._self_to_m_cache[2] )
            [([1, 1], [([1, 1], t + 1), ([2], 1)]), ([2], [([1, 1], 1), ([2], 1)])]
            sage: l( HCosp3._m_to_self_cache[2] )
            [([1, 1], [([1, 1], 1/t), ([2], -1/t)]),
             ([2], [([1, 1], -1/t), ([2], (t + 1)/t)])]
        """
        self._invert_morphism(n, QQt, self._self_to_m_cache,
                              self._m_to_self_cache,
                              to_other_function=self._to_m)

    class Element(sfa.SymmetricFunctionAlgebra_generic.Element):
        pass


# the H-spin basis
class LLT_spin(LLT_generic):

    def __init__(self, llt):
        r"""
        A class of methods for the h-spin LLT basis of the symmetric functions.

        INPUT:

        - ``self`` -- an instance of the LLT hcospin basis
        - ``llt`` -- a family of LLT symmetric function bases

        TESTS::

            sage: HSp3 = SymmetricFunctions(FractionField(QQ['t'])).llt(3).hspin()
            sage: TestSuite(HSp3).run(skip = ["_test_associativity", "_test_distributivity", "_test_prod"]) # products are too expensive, long time (10s on sage.math, 2012)
            sage: TestSuite(HSp3).run(elements = [HSp3.t*HSp3[1,1]+HSp3.t*HSp3[2], HSp3[1]+(1+HSp3.t)*HSp3[1,1]])  # long time (depends on previous)

        ::

            sage: HS3t2 = SymmetricFunctions(QQ).llt(3,t=2).hspin()
            sage: TestSuite(HS3t2).run() # products are too expensive, long time (7s on sage.math, 2012)

        ::

            sage: R.<x> = QQ[]
            sage: HS3x = SymmetricFunctions(FractionField(R)).llt(3, t=x).hspin()
            sage: TestSuite(HS3x).run(skip = ["_test_associativity", "_test_distributivity", "_test_prod"]) # products are too expensive, long time (4s on sage.math, 2012)
            sage: TestSuite(HS3x).run(elements = [HS3x.t*HS3x[1,1]+HS3x.t*HS3x[2], HS3x[1]+(1+HS3x.t)*HS3x[1,1]])  # long time (depends on previous)
        """
        level = llt._k
        if level not in hsp_to_m_cache:
            hsp_to_m_cache[level] = {}
            m_to_hsp_cache[level] = {}
        self._self_to_m_cache = hsp_to_m_cache[level]
        self._m_to_self_cache = m_to_hsp_cache[level]

        LLT_generic.__init__(self, llt, prefix="HSp%s" % level)

    def _to_m(self, part):
        r"""
        Return a function which gives the coefficient of a partition
        in the monomial expansion of self(part).

        INPUT:

        - ``self`` -- an instance of the LLT hspin basis
        - ``part`` -- a partition

        OUTPUT:

        - returns a function which accepts a partition and returns the coefficient
          in the expansion of the monomial basis

        EXAMPLES::

            sage: # needs sage.symbolic
            sage: HSp3 = SymmetricFunctions(FractionField(QQ['t'])).llt(3).hspin()
            sage: f21 = HSp3._to_m(Partition([2,1]))
            sage: [f21(p) for p in Partitions(3)]
            [t, t + 1, t + 2]
            sage: HSp3.symmetric_function_ring().m()( HSp3[2,1] )
            (t+2)*m[1, 1, 1] + (t+1)*m[2, 1] + t*m[3]
        """
        level = self.level()
        f = lambda part2: QQt(ribbon_tableau.spin_polynomial([level*i for i in part], part2, level))
        return f

    class Element(LLT_generic.Element):
        pass


# the h-cospin basis
class LLT_cospin(LLT_generic):
    def __init__(self, llt):
        r"""
        A class of methods for the h-cospin LLT basis of the symmetric functions.

        INPUT:

        - ``self`` -- an instance of the LLT hcospin basis
        - ``llt`` -- a family of LLT symmetric function bases

        TESTS::

            sage: HCosp3 = SymmetricFunctions(FractionField(QQ['t'])).llt(3).hcospin()
            sage: TestSuite(HCosp3).run(skip = ["_test_associativity", "_test_distributivity", "_test_prod"]) # products are too expensive, long time (11s on sage.math, 2012)
            sage: TestSuite(HCosp3).run(elements = [HCosp3.t*HCosp3[1,1]+HCosp3.t*HCosp3[2], HCosp3[1]+(1+HCosp3.t)*HCosp3[1,1]])  # long time (depends on previous)

        ::

            sage: HC3t2 = SymmetricFunctions(QQ).llt(3,t=2).hcospin()
            sage: TestSuite(HC3t2).run() # products are too expensive, long time (6s on sage.math, 2012)

        ::

            sage: R.<x> = QQ[]
            sage: HC3x = SymmetricFunctions(FractionField(R)).llt(3, t=x).hcospin()
            sage: TestSuite(HC3x).run(skip = ["_test_associativity", "_test_distributivity", "_test_prod"]) # products are too expensive, long time (5s on sage.math, 2012)
            sage: TestSuite(HC3x).run(elements = [HC3x.t*HC3x[1,1]+HC3x.t*HC3x[2], HC3x[1]+(1+HC3x.t)*HC3x[1,1]])  # long time (depends on previous)
        """
        level = llt._k
        if level not in hcosp_to_m_cache:
            hcosp_to_m_cache[level] = {}
            m_to_hcosp_cache[level] = {}
        self._self_to_m_cache = hcosp_to_m_cache[level]
        self._m_to_self_cache = m_to_hcosp_cache[level]
        LLT_generic.__init__(self, llt, prefix="HCosp%s" % level)

    def _to_m(self, part):
        r"""
        Return a function which gives the coefficient of part2 in the
        monomial expansion of self(part).

        INPUT:

        - ``self`` -- an instance of the LLT hcospin basis
        - ``part`` -- a partition

        OUTPUT:

        - returns a function which accepts a partition and returns the coefficient
          in the expansion of the monomial basis

        EXAMPLES::

            sage: HCosp3 = SymmetricFunctions(FractionField(QQ['t'])).llt(3).hcospin()
            sage: f21 = HCosp3._to_m(Partition([2,1]))
            sage: [f21(p) for p in Partitions(3)]
            [1, t + 1, 2*t + 1]
            sage: HCosp3.symmetric_function_ring().m()( HCosp3[2,1] )
            (2*t+1)*m[1, 1, 1] + (t+1)*m[2, 1] + m[3]
        """
        level = self.level()
        f = lambda part2: QQt(ribbon_tableau.cospin_polynomial([level*i for i in part], part2, level))
        return f

    class Element(LLT_generic.Element):
        pass


# Backward compatibility for unpickling
from sage.misc.persist import register_unpickle_override

register_unpickle_override('sage.combinat.sf.llt', 'LLTElement_spin', LLT_spin.Element)
register_unpickle_override('sage.combinat.sf.llt', 'LLTElement_cospin', LLT_cospin.Element)



# sage_setup: distribution = sagemath-combinat
# sage.doctest: needs sage.groups
"""
Miscellaneous functions
"""
# ****************************************************************************
#       Copyright (C) 2008 Mike Hansen <mhansen@gmail.com>,
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
# ****************************************************************************

from functools import wraps

from sage.misc.misc_c import prod
from sage.misc.lazy_import import lazy_import

lazy_import('sage.groups.perm_gps.permgroup', ['PermutationGroup', 'PermutationGroup_generic'])
lazy_import('sage.groups.perm_gps.constructor', 'PermutationGroupElement')
lazy_import('sage.groups.perm_gps.permgroup_named', 'SymmetricGroup')


def change_support(perm, support, change_perm=None):
    """
    Changes the support of a permutation defined on [1, ..., n] to
    support.

    EXAMPLES::

        sage: from sage.combinat.species.misc import change_support
        sage: p = PermutationGroupElement((1,2,3)); p
        (1,2,3)
        sage: change_support(p, [3,4,5])
        (3,4,5)
    """
    if change_perm is None:
        change_perm = prod([PermutationGroupElement((i+1, support[i]))
                            for i in range(len(support))
                            if i+1 != support[i]],
                           PermutationGroupElement([], SymmetricGroup(support)))

    if isinstance(perm, PermutationGroup_generic):
        return PermutationGroup([change_support(g, support, change_perm) for g in perm.gens()])

    return change_perm*perm*~change_perm


def accept_size(f):
    """
    The purpose of this decorator is to change calls like
    species.SetSpecies(size=1) to species.SetSpecies(min=1, max=2).
    This is to make caching species easier and to restrict the number
    of parameters that the lower level code needs to know about.

    EXAMPLES::

        sage: from sage.combinat.species.misc import accept_size
        sage: def f(*args, **kwds):
        ....:       print("{} {}".format(args, sorted(kwds.items())))
        sage: f = accept_size(f)
        sage: f(min=1)
        () [('min', 1)]
        sage: f(size=2)
        () [('max', 3), ('min', 2)]
    """
    @wraps(f)
    def new_func(*args, **kwds):
        if 'size' in kwds:
            if 'min' in kwds or 'max' in kwds:
                raise ValueError("cannot specify both size and (min or max)")
            kwds['min'] = kwds['size']
            kwds['max'] = kwds['size'] + 1
            del kwds['size']
        return f(*args, **kwds)
    return new_func



# sage_setup: distribution = sagemath-combinat
r"""
Combinatorics quickref

Integer Sequences::

    sage: s = oeis([1,3,19,211]); s                         # optional - internet
    0: A000275: ...
    sage: s[0].programs()                                   # optional - internet
    [('maple', ...),
     ('mathematica', ...),
     ('pari',
      0: {a(n) = if( n<0, 0, n!^2 * 4^n * polcoeff( 1 / besselj(0, x + x * O(x^(2*n))), 2*n))}; /* _Michael Somos_, May 17 2004 */)]

Combinatorial objects::

    sage: S = Subsets([1,2,3,4]); S.list(); S.<tab>                       # not tested
    sage: P = Partitions(10000); P.cardinality()                                        # needs sage.libs.flint
    3616...315650422081868605887952568754066420592310556052906916435144
    sage: Combinations([1,3,7]).random_element()            # random
    sage: Compositions(5, max_part=3).unrank(3)
    [2, 2, 1]

    sage: DyckWord([1,0,1,0,1,1,0,0]).to_binary_tree()                                  # needs sage.graphs
    [., [., [[., .], .]]]
    sage: Permutation([3,1,4,2]).robinson_schensted()
    [[[1, 2], [3, 4]], [[1, 3], [2, 4]]]
    sage: StandardTableau([[1, 4], [2, 5], [3]]).schuetzenberger_involution()
    [[1, 3], [2, 4], [5]]

Constructions and Species::

    sage: for (p, s) in cartesian_product([P,S]): print((p, s)) # not tested
    sage: def IV_3(n):
    ....:     return IntegerVectors(n, 3)
    sage: DisjointUnionEnumeratedSets(Family(IV_3, NonNegativeIntegers))  # not tested

Words::

    sage: Words('abc', 4).list()
    [word: aaaa, ..., word: cccc]

    sage: Word('aabcacbaa').is_palindrome()
    True
    sage: WordMorphism('a->ab,b->a').fixed_point('a')
    word: abaababaabaababaababaabaababaabaababaaba...

Polytopes::

    sage: points = random_matrix(ZZ, 6, 3, x=7).rows()                                  # needs sage.modules
    sage: L = LatticePolytope(points)                                                   # needs sage.geometry.polyhedron sage.modules
    sage: L.npoints(); L.plot3d()                           # random                    # needs palp sage.geometry.polyhedron sage.modules sage.plot

:ref:`Root systems, Coxeter and Weyl groups <sage.combinat.root_system.all>`::

    sage: WeylGroup(["B",3]).bruhat_poset()                                             # needs sage.graphs sage.modules
    Finite poset containing 48 elements
    sage: RootSystem(["A",2,1]).weight_lattice().plot()         # not tested            # needs sage.graphs sage.modules sage.plot

:ref:`Crystals <sage.combinat.crystals.all>`::

    sage: CrystalOfTableaux(["A",3], shape=[3,2]).some_flashy_feature()   # not tested

:mod:`Symmetric functions and combinatorial Hopf algebras <sage.combinat.algebraic_combinatorics>`::

    sage: Sym = SymmetricFunctions(QQ); Sym.inject_shorthands(verbose=False)            # needs sage.sage.modules
    sage: m( ( h[2,1] * (1 + 3 * p[2,1]) ) + s[2](s[3]) )                               # needs sage.sage.modules
    3*m[1, 1, 1] + ... + 10*m[5, 1] + 4*m[6]

:ref:`Discrete groups, Permutation groups <sage.groups.groups_catalog>`::

    sage: S = SymmetricGroup(4)                                                         # needs sage.groups
    sage: M = PolynomialRing(QQ, 'x0,x1,x2,x3')
    sage: M.an_element() * S.an_element()                                               # needs sage.groups
    x0

Graph theory, posets, lattices (:ref:`sage.graphs`, :ref:`sage.combinat.posets.all`)::

    sage: Poset({1: [2,3], 2: [4], 3: [4]}).linear_extensions().cardinality()           # needs sage.graphs sage.modules
    2
"""
